#include "point.h"


point::point(){
}

point::point(int numDimensions, float* po, bool inMicro){
	this->numDimensions = numDimensions;
	this->p = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		this->p[i] = po[i];
	}
	this->inMicro = inMicro;
	isAssigned = false;
}

point::point(const point& po){
	numDimensions = po.numDimensions;
	p = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		p[i] = po.p[i];
	}
	inMicro = po.inMicro;
	isAssigned = po.isAssigned;
}

point& point::operator= (const point& po){
	this->numDimensions = po.numDimensions;
	this->p = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		this->p[i] = po.p[i];
	}
	this->inMicro = po.inMicro;
	this->isAssigned = po.isAssigned;

	return *this;
}


void point::calculateHilScore(float** extent){
	hilScore = new HilScore(p, extent, numDimensions, 4);
	cout << *this << " hilScore=" << *hilScore << endl;
}

bool point::operator==(const point& po) const{
	bool isit = true;
	if (numDimensions!=po.numDimensions){
		cerr << "comparing points of different dimensionality" << endl;
		return false;
	}
	for (int i=0; i<numDimensions; i++){
		if (abs(p[i]-po.p[i])>EPSILON){
			isit = false;
		}
	}
	return isit;
}


point::~point(){
	delete[] p;
}


float** newExtent(float** oldExtent, pointPtr po){
	int numDimensions = po->numDimensions;
	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}

	for (int d=0; d<numDimensions; d++){
		extent[d][0] = oldExtent[d][0];
		extent[d][1] = oldExtent[d][1];
		if (po->p[d] < oldExtent[d][0]){
			extent[d][0] = po->p[d];
		}
		if (po->p[d] > oldExtent[d][1]){
			extent[d][1] = po->p[d];
		}
	}
	return extent;
}



float** findExtent(list<pointPtr> poiset){
	int numDimensions = poiset.front()->numDimensions;

	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}

	list<pointPtr>::iterator iter;

	for (iter=poiset.begin(); iter!=poiset.end(); iter++){
		for (int d=0; d<numDimensions; d++){
			if (iter==poiset.begin()){
				extent[d][0] = (*iter)->p[d];
				extent[d][1] = (*iter)->p[d];
			}
			else{
				if ((*iter)->p[d] < extent[d][0]){
					extent[d][0] = (*iter)->p[d];
				}
				if ((*iter)->p[d] > extent[d][1]){
					extent[d][1] = (*iter)->p[d];
				}
			}
		}
	}

	return extent;
}


float** getUnion(float** extentA, float** extentB, int numDimensions){
	float** theunion = new float*[numDimensions];
	for (int d=0; d<numDimensions; d++){
		theunion[d] = new float[2];
		theunion[d][0] = min(extentA[d][0], extentB[d][0]);
		theunion[d][1] = max(extentA[d][1], extentB[d][1]);
	}
	return theunion;
}


float getNormalizedPerimeter(float** extent, float* sizeWorld, int numDimensions){
	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
	}
	return perimeter;
}


float** findExtent(vector<pointPtr> poiset){
	int numDimensions = poiset[0]->numDimensions;

	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}

	for (uint i=0; i<poiset.size(); i++){
		for (int d=0; d<numDimensions; d++){
			if (i==0){
				extent[d][0] = poiset[i]->p[d];
				extent[d][1] = poiset[i]->p[d];
			}
			else{
				if (poiset[i]->p[d] < extent[d][0]){
					extent[d][0] = poiset[i]->p[d];
				}
				if (poiset[i]->p[d] > extent[d][1]){
					extent[d][1] = poiset[i]->p[d];
				}
			}
		}
	}

	return extent;
}

float** findExtentPadded(vector<pointPtr> poiset){
	int numDimensions = poiset[0]->numDimensions;

	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}

	for (uint i=0; i<poiset.size(); i++){
		for (int d=0; d<numDimensions; d++){
			if (i==0){
				extent[d][0] = poiset[i]->p[d];
				extent[d][1] = poiset[i]->p[d];
			}
			else{
				if (poiset[i]->p[d]<extent[d][0]){
					extent[d][0] = poiset[i]->p[d];
				}
				if (poiset[i]->p[d]>extent[d][1]){
					extent[d][1] = poiset[i]->p[d];
				}
			}
		}
	}
	for (int d=0; d<numDimensions; d++){
		extent[d][0] = extent[d][0] - 1;
		extent[d][1] = extent[d][1] + 1;
	}

	return extent;
}


vector<pointPtr> SortMerge(vector<pointPtr> publicPoints, vector<pointPtr> microPoints){
	vector<pointPtr> result;
//	int numDimensions = publicPoints[0]->numDimensions;

//	float** extent = findExtentPadded(publicPoints);

//	for (uint i=0; i<publicPoints.size(); i++){
//		publicPoints[i]->calculateHilScore(extent);
//	}
//
//	for (uint i=0; i<microPoints.size(); i++){
//		microPoints[i]->calculateHilScore(extent);
//	}

	sort(publicPoints.begin(), publicPoints.end(), cmpPointPtr);
	sort(microPoints.begin(), microPoints.end(), cmpPointPtr);

	vector<pointPtr>::iterator pubiter = publicPoints.begin();
	vector<pointPtr>::iterator miciter = microPoints.begin();

	while (pubiter!=publicPoints.end()){
//		cout << ".";
		if (miciter==microPoints.end()){
			if ( !(*pubiter)->isAssigned ){
				result.push_back(*pubiter);
				(*pubiter)->isAssigned = true;
			}
			pubiter++;
		}
		else{
//			cout << "checking pub:" << *(*pubiter) << " with mic:" << *(*miciter) << endl;
			if ( *(*miciter) < *(*pubiter) ){
//				cout << " mic < pub" << endl;
				miciter++;
			}
			else if ( *(*miciter) == *(*pubiter) ){
//				cout << " mic == pub" << endl;
				result.push_back(*miciter);
				(*pubiter)->isAssigned = true;
				miciter++;
				pubiter++;
			}
			else{// if ( *(*pubiter) < *(*miciter) ){
//				cout << " mic > pub" << endl;
				if ( !(*pubiter)->isAssigned ){
					result.push_back(*pubiter);
					(*pubiter)->isAssigned = true;
				}
				pubiter++;
			}
		}
	}
//	cout << endl;

//	publicPoints.clear();
//	microPoints.clear();

//	for (int d=0; d<numDimensions; d++){
//		delete[] extent[d];
//	}
//	delete[] extent;

	return result;
}


vector<pointPtr> readFromFile(char* filename, bool isMicro, int &numDimensions){
	if (isMicro){
		cerr << "reading microData from " << filename << endl;
	}
	else{
		cerr << "reading publicData from " << filename << endl;
	}
    vector<pointPtr> poiset;
    // read from publicData
    fstream fin(filename, ios::in);
	fin >> numDimensions; // first read dimensionality
	float* p = new float[numDimensions];
	while(!fin.eof()){
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		pointPtr po = new point(numDimensions, p, isMicro);
		poiset.push_back(po);
	}
	fin.close();
	cerr << "... done reading " << poiset.size() << " points" << endl;
	delete p;
	return poiset;
}


bool operator<(const point& lhs, const point& rhs){
	for (int d=0; d<lhs.numDimensions; d++){
		if (lhs.p[d]==rhs.p[d]){
			continue;
		}
		else if (lhs.p[d]<rhs.p[d]){
			return true;
		}
		else{
			return false;
		}
	}
	return false;

//	cout << "operator<:: lhs=" << *(lhs.hilScore) << " rhs=" << *(rhs.hilScore) << endl;
//	return ( *(lhs.hilScore)<*(rhs.hilScore) );
}

bool cmpPointPtr(const pointPtr& lhs, const pointPtr& rhs){
	for (int d=0; d<lhs->numDimensions; d++){
		if (lhs->p[d]==rhs->p[d]){
			continue;
		}
		else if (lhs->p[d]<rhs->p[d]){
			return true;
		}
		else{
			return false;
		}
	}
	return false;

//	cout << "cmpPointPtr:: lhs=" << *(lhs->hilScore) << " rhs=" << *(rhs->hilScore) << endl;
//	return ( *(lhs->hilScore)<*(rhs->hilScore) );
}



ostream& operator<<(ostream& os, const point& po){
	os << "(" << po.p[0];
	for (int i=1; i<po.numDimensions; i++){
		os << "," << po.p[i];
	}
	os << ")" << "|" << po.inMicro;
	return os;
}


string getString(float* point, int numDimensions){
	ostringstream sstream;

	sstream << "(" << point[0];
	for (int i=1; i<numDimensions; i++){
		sstream << "," << point[i];
	}
	sstream << ")";
	return sstream.str();
}

string getString(float** rect, int numDimensions){
	ostringstream sstream;

	float* lower = new float[numDimensions];
	float* upper = new float[numDimensions];

	for (int i=0; i<numDimensions; i++){
		lower[i] = rect[i][0];
		upper[i] = rect[i][1];
	}
	sstream << getString(lower, numDimensions) << "--" << getString(upper, numDimensions);

	delete[] lower;
	delete[] upper;

	return sstream.str();
}


bool inside(float* point, float ** rect, int numDimensions){
	bool isit = true;
	for (int i=0; i<numDimensions; i++){
		if ( point[i]<rect[i][0] || point[i]>rect[i][1] ){
			isit = false;
		}
	}
	return isit;
}

bool insideStr(float* point, float ** rect, int numDimensions){
	bool isit = true;
	for (int i=0; i<numDimensions; i++){
		if ( point[i]<=rect[i][0] || point[i]>rect[i][1] ){
			isit = false;
		}
	}
	return isit;
}



float** intersect(float **rect1, float **rect2, int numDimensions){
	float **inter = new float*[numDimensions]; // box intersection

	for (int i=0; i<numDimensions; i++){
		inter[i] = new float[2];
		inter[i][0] = max(rect1[i][0], rect2[i][0]); // lower point in dim
		inter[i][1] = min(rect1[i][1], rect2[i][1]); // higher point in dim
	}

	return inter;
}


bool overlaps(float **rect1, float **rect2, int numDimensions){
	bool isit = true;

//	cout << "rect1:: " << getString(rect1, numDimensions) << endl;
//	cout << "rect2:: " << getString(rect2, numDimensions) << endl;

	float **inter = new float*[numDimensions]; // box intersection

	for (int i=0; i<numDimensions; i++){
		inter[i] = new float[2];
		inter[i][0] = max(rect1[i][0], rect2[i][0]); // lower point in dim
		inter[i][1] = min(rect1[i][1], rect2[i][1]); // higher point in dim
		if (inter[i][0]>inter[i][1]){
			isit = false;
		}
	}

	for (int i=0; i<numDimensions; i++){
		delete[] inter[i];
	}
	delete[] inter;

//	cout << "overlaps = " << isit << endl;
	return isit;
}

bool overlapsStr(float **rect1, float **rect2, int numDimensions){
	bool isit = true;

//	cout << "rect1:: " << getString(rect1, numDimensions) << endl;
//	cout << "rect2:: " << getString(rect2, numDimensions) << endl;

	float **inter = new float*[numDimensions]; // box intersection

	for (int i=0; i<numDimensions; i++){
		inter[i] = new float[2];
		inter[i][0] = max(rect1[i][0], rect2[i][0]); // lower point in dim
		inter[i][1] = min(rect1[i][1], rect2[i][1]); // higher point in dim
		if (inter[i][0]>=inter[i][1]){
			isit = false;
		}
	}

	for (int i=0; i<numDimensions; i++){
		delete[] inter[i];
	}
	delete[] inter;

//	cout << "overlapsStr = " << isit << endl;
	return isit;
}



