#ifndef POINT_H_
#define POINT_H_

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <vector>
#include <list>
#include <assert.h>
#include <algorithm>


#include "../hilbert/hilScore.h"


typedef unsigned int uint;

#define EPSILON 0.00001

using namespace std;

typedef class point* pointPtr;

class point
{
public:
	int numDimensions;
	bool inMicro;
	bool isAssigned; // FIX fot Topdown, so that a point is not assigned to more than one group
	float* p;
	HilScore *hilScore;

	point();
	point(int numDimensions, float* p, bool inMicro);
	point(const point& po); // copy constructor
	~point();

	void calculateHilScore(float** extent);

	bool operator==(const point& po) const;
	point& operator= (const point& po);

	friend ostream& operator<<(ostream& os, const point& po);

};

bool operator<(const point& lhs, const point& rhs);
bool cmpPointPtr(const pointPtr& lhs, const pointPtr& rhs);

ostream& operator<<(ostream& os, const point& po);


// functions to manipulate vector<pointPtr>
float** newExtent(float** oldExtent, pointPtr po);
float** findExtent(vector<pointPtr> poiset);
float** findExtent(list<pointPtr> poiset);
vector<pointPtr> SortMerge(vector<pointPtr> publicPoints, vector<pointPtr> microPoints);
vector<pointPtr> readFromFile(char* filename, bool isMicro, int &numDimensions);


string getString(float* point, int numDimensions);
string getString(float** rect, int numDimensions);


float** getUnion(float** extentA, float** extentB, int numDimensions);
float getNormalizedPerimeter(float** extent, float* sizeWorld, int numDimensions);


bool insideStr(float* point, float ** rect, int numDimensions);
bool inside(float* point, float ** rect, int numDimensions);


float** intersect(float **rect1, float **rect2, int numDimensions);


// functor for comparing pointPtr by any dimension
class orderByDim{
public:
	int dim;
	orderByDim(int d){
		dim = d;
	}
	bool operator()(const pointPtr& a, const pointPtr& b) const{
		if (a->p[dim]<b->p[dim]){
			return true;
		}
		return false;
	}
};



#endif /*POINT_H_*/
