#include "generator.h"

using namespace std;

int main(int argc, char **argv){
	
	if (argc!=5){
		cout << "Usage: " << argv[0] << " [numDimensions] [numObjects] [dimSize] [skew]" << endl;
		exit(0);
	}
		
	
	int numDimensions = atoi(argv[1]);
	int numObjects = atoi(argv[2]);
	int dimSize = atoi(argv[3]);
	float skew = atof(argv[4]);
	
	ostringstream filename;
	filename << "d" << numDimensions << "_" << dimSize << "_t" << numObjects << "_z" << skew;
	
	ofstream fout;
	fout.open (filename.str().c_str(), ofstream::out);
	
	float *probs = new float[dimSize];
	
	float sum = 0;
	for (int i=0; i<dimSize; i++){
		probs[i] = 1/pow((float) i+1, skew);
		sum += probs[i];
	}
	
	// normalize so that the sum is 1
	for (int i=0; i<dimSize; i++){
		probs[i] = probs[i]/sum;
	}
	
	vector<int>* permutations = new vector<int>[numDimensions];
	for (int i=0; i<dimSize; i++){
		for (int j=0; j<numDimensions; j++){
			permutations[j].push_back(i);
		}
	}
	for (int j=0; j<numDimensions; j++){
		random_shuffle(permutations[j].begin(), permutations[j].end());
	}
	
	
	srand48(123456); // random seed
	
	fout << numDimensions << endl;
	
	for (int t=0; t<numObjects; t++){
		if (t>0) fout << endl;
		for (int d=0; d<numDimensions; d++){
			double rnd = drand48();
			int index = 0;
			float sum = probs[0];
			while(sum<rnd && index<dimSize){
				sum += probs[index];
				index++;
			}
			if (index==dimSize) index--;
			if (d==0){
				fout << permutations[d][index];
			}
			else{
				fout << " " << permutations[d][index];
			}
		}		
	}
	

	
}
