#include "hilKAnomPD.h"



HilKAnomPD::HilKAnomPD(char* microTable, char* publicData){
	groupsCalculated = false;
	
	// micro Table
	cerr << "HilKAnomPD:: reading from microTable " << microTable << endl;
	ifstream fin;
	fin.open(microTable);
	
	fin >> numDimensions; // first read dimensionality
	float p[numDimensions];
	
	vector<point> microPoints;
	while(!fin.eof()){
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		point* po = new point(numDimensions, p, true);
		microPoints.push_back(*po);
	}
	fin.close();
	cerr << "...done reading " << microPoints.size() << " points" << endl;
	
	// public Data
	cerr << "HilKAnomPD:: reading from publicData " << publicData << endl;
	ifstream fin2;
	fin2.open(publicData);
	
	fin2 >> numDimensions; // first read dimensionality
//	cout << "numDimensions=" << numDimensions << endl;
	
	float* mini = new float[numDimensions];
	float* maxi = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		mini[i] = FLT_MAX;
		maxi[i] = FLT_MIN;
	}
	
	while(!fin2.eof()){
		for (int i=0; i<numDimensions; i++){
			fin2 >> p[i];
			if (p[i]<mini[i]){
				mini[i] = p[i];
			}
			if (p[i]>maxi[i]){
				maxi[i] = p[i];
			}
		}
		point* po = new point(numDimensions, p, false);
		
		vector<point>::iterator iter;
		bool inMicro = false;
		for (iter=microPoints.begin(); iter!=microPoints.end(); iter++){
			if (*iter==*po){
				inMicro = true;
				break;
			}
		}
		if (inMicro){
//			cout << "*";
			po->inMicro = true;
		}
//		else{
//			cout << "-";
//		}
		
		points.push_back(*po);
	}
	
	fin2.close();
	
	
	
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = maxi[i] - mini[i];
	}
	delete[] mini;
	delete[] maxi;
	
	cerr << "done reading " << points.size() << " points" << endl;
	
	kdtree = new KDTree(publicData, 1); // build a KDTree...
//	cout << "KDTree built" << endl;
	
}


HilKAnomPD::~HilKAnomPD(){
	delete[] sizeWorld;
	delete kdtree;
}


float* HilKAnomPD::getSizeWorld(){
	return sizeWorld;
}


vector<group> HilKAnomPD::getGroups(int K){
	if (!groupsCalculated || this->K!=K){
		groups.clear();
		
		groupsCalculated = true;
		this->K = K;
		
		vector<point>::iterator iter;
		iter = points.begin();
		
		bool groupOK = false;
		while(iter!=points.end()){
			int count = 0;
			int microCount = 0;
			group* gp = new group(numDimensions);
			
			groupOK = false;
			while(!groupOK && iter!=points.end()){
				gp->insertPoint(*iter);
				if (iter->inMicro){
					microCount++;
				}
				count++;
				iter++;
				if (microCount>0 && count>=K){
					groupOK = true;
				}
			}
//			cout << "count=" << count << " microCount=" << microCount << endl;
			
			if (groupOK){
				float** extent = gp->getExtent();
				vector<pointPtr> result = kdtree->rangeQuery(extent);
				for (uint j=0; j<result.size(); j++){
					gp->insertPoint(*(result[j]));
				}
				groups.push_back(*gp);	
			}
			else{ // merge this with last group
				if (groups.size()==0){
					cerr << "cannot create any group" << endl;
					return groups; // exit with no group
				}
				group lastGroup = groups[groups.size()-1];
				groups.pop_back();
				set<point>::iterator it;
				for (it=gp->points.begin(); it!=gp->points.end(); it++){
					point po = *it;
					lastGroup.insertPoint(po);
				}
				float** extent = lastGroup.getExtent();
				vector<pointPtr> result = kdtree->rangeQuery(extent);
				for (uint j=0; j<result.size(); j++){
					lastGroup.insertPoint(*(result[j]));
				}
				groups.push_back(lastGroup);
			}
			
		}
	}
	return groups;
}


