#include "kdnode.h"


KDNode::KDNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension){
	this->numDimensions = numDimensions;
	this->numPoints = numPoints;
	this->points = points;
	this->splitDimension = splitDimension;
	this->numMicroPoints = calculateNumMicroPoints();
	splitValue = -2007;
	left = NULL;
	right = NULL;
	parent = NULL;
	leafNode = true;
	boostedLeafNode = false;
	leafTuples = 0;
	perceivedMicroTuples = 0;
}

KDNode::~KDNode(){
//	cerr << "called ~KDNode()" << endl;
	for (int i=0; i<numDimensions; i++){
		delete[] extent[i];
		delete[] mbr[i];
	}
	delete[] extent;
	delete[] mbr;
//	for (int i=0; i<numPoints; i++){
//		delete points[i];
//	}
	delete[] points;
}

int KDNode::calculateNumMicroPoints(){
	numMicroPoints = 0;
	for (int i=0; i<numPoints; i++){
		if (points[i]->inMicro){
			numMicroPoints++;
		}
	}
	return numMicroPoints;
}

void KDNode::calculateExtentMBR(float pad){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}

		vector<float>::iterator it = min_element(temp.begin(), temp.end());

		mbr[d][0] = *min_element(temp.begin(), temp.end()) - pad;
		mbr[d][1] = *max_element(temp.begin(), temp.end()) + pad;

		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}

		temp.clear();
	}
}


void KDNode::calculateExtentMBR(){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}

		vector<float>::iterator it = min_element(temp.begin(), temp.end());

		mbr[d][0] = *min_element(temp.begin(), temp.end());
		mbr[d][1] = *max_element(temp.begin(), temp.end());

		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}

		temp.clear();
	}
}


bool KDNode::splitNode(KDNodePtr &left, KDNodePtr &right){
//	cout << "called splitNode for id=" << this->id << endl;
//	cout << "splitDimension = " << splitDimension << endl;
	vector<float> temp;
	for (int i=0; i<numPoints; i++){
		temp.push_back(points[i]->p[splitDimension]);
	}

	sort(temp.begin(), temp.end()); // sort the vector!

	float median = *(temp.begin()+(temp.size()-1)/2);
	splitValue = median;



//	/*
	// *******************************************
	// this code segment splits in middle ********
	int leftNumPoints = temp.size()/2 +  temp.size()%2;
	int rightNumPoints = temp.size() - leftNumPoints;
//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;

	pointPtr *leftPoints, *rightPoints;

	leftPoints = new pointPtr [leftNumPoints];
	rightPoints = new pointPtr [rightNumPoints];

	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=median && leftIndex<leftNumPoints){ // store them in left child
//			cout << "left" << endl;
			leftPoints[leftIndex] = new point(*(points[i]));
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			rightPoints[rightIndex] = new point(*(points[i]));
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	// *******************************************
//	*/

	/*
	// *******************************************
	// this code segment splits in median ******
	int leftNumPoints = (temp.size()-1)/2;
	for (int i=leftNumPoints; i<numPoints; i++){
		if ( fabs(temp[i]-median)<0.001 ){
			leftNumPoints++;
		}
		else break;
	}

	int rightNumPoints = temp.size() - leftNumPoints;
//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;

	pointPtr *leftPoints, *rightPoints;

	leftPoints = new pointPtr [leftNumPoints];
	rightPoints = new pointPtr [rightNumPoints];

	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=median){ // store them in left child
//			cout << "left" << endl;
//			leftPoints[leftIndex] = new point(*(points[i]));
			leftPoints[leftIndex] = points[i];
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
//			rightPoints[rightIndex] = new point(*(points[i]));
			rightPoints[rightIndex] = points[i];
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	// *******************************************
	*/

	if (rightNumPoints==0){ // don't split
		delete[] leftPoints;
		delete[] rightPoints;
//		cerr << "**** rightNumPoints==0" << endl;
		return false; // don't split
	}

	this->leafNode = false; // make this a non-leaf node!

	left = new KDNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new KDNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;

	this->left = left;
	this->right = right;

	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;

	temp.clear();

	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;

	return true;
}

void KDNode::printNode(){
	cout << this;
}

ostream& operator<<(ostream& os, const KDNodePtr& n){
	os << "node ID=" << n->id << " leafNode=" << n->leafNode << " ";
	os << "numDimensions=" << n->numDimensions << " numPoints=" << n->numPoints << " numMicroPoints=" << n->numMicroPoints << endl;
	os << "splitDimension=" << n->splitDimension << " splitValue=" << n->splitValue << endl;
	for (int i=0; i<n->numPoints; i++){
		os << *(n->points[i]) << " ";
	}
	os << endl;
	os << "mbr:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->mbr[j][0] << " -- " << n->mbr[j][1] << " | ";
	}
	os << endl;
	os << "extent:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->extent[j][0] << " -- " << n->extent[j][1] << " | ";
	}
	os << endl;

	if (n->left==NULL) cout << "left node is null" << endl;
	if (n->right==NULL) cout << "right node is null" << endl;

	return os;
}

