#ifndef KDNODE_H_
#define KDNODE_H_

#include <algorithm>
#include <vector>
#include <iostream>
#include <cmath>
#include "../common/point.h"

using namespace std;


typedef class KDNode *KDNodePtr;

class KDNode
{
public:
int id; // node id
bool leafNode;
bool boostedLeafNode;
int numDimensions; // number of dimensions
int numPoints; // number of points
int numMicroPoints; // number of points from microdata
pointPtr *points; // the array of points
float **extent; // the extent in low-high pairs
float **mbr; // the mbr of the points in low-high pairs
int splitDimension; // the dimension along which children are split
float splitValue; // the value at which children are split
double leafTuples; // the number of all tuples under this node
double perceivedMicroTuples; // the perceived number volume of all MT tuples at this leaf; only for leaves

KDNodePtr left, right, parent; // pointers to children and parent nodes



KDNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension); // constructor

~KDNode();

int calculateNumMicroPoints();

void calculateExtentMBR();
void calculateExtentMBR(float pad); // add pad around extend and mbr, for root node only
bool splitNode(KDNodePtr &left, KDNodePtr &right); // split and create children

void printNode();

friend ostream& operator<<(ostream& os, const KDNodePtr& t);
};

ostream& operator<<(ostream& os, const KDNodePtr& n);

#endif /*KDNODE_H_*/
