#ifndef KDTREE_H_
#define KDTREE_H_

#include <fstream>
#include <sstream>
#include <map>
#include <set>
#include <string>
#include <algorithm>
#include "kdnode.h"
#include "../common/group.h"
#include "../common/Timing.h"
//#include "../minpertree/minpertree.h"

using namespace std;


typedef class KDTree *KDTreePtr;

class KDTree
{
public:
	int numDimensions; // dimensionality

	map<int, KDNodePtr> nodes;
//	vector<NodePtr> nodes;

	int K; // anonymity requirement

	vector<group> groups; // K-anonymized groups

	ofstream fout;
	KDNodePtr root;

	float* sizeWorld;

	KDTree(char *inputfile, int K);
	KDTree(vector<pointPtr> poiset, int K);

	void recursiveSplit(KDNodePtr node);

//	float* getSizeWorld();
	void calculateSizeWorld();

	void writeGroups(char *inputfile);
	vector<group> getGroups();
	void printMetrics();

	vector<pointPtr> rangeQuery(float **query);
	void recRangeQuery(KDNodePtr node, float **query, vector<pointPtr>& res); // recursive descend

	// strict versions:
	vector<pointPtr> rangeQueryStr(float **query);
	void recRangeQueryStr(KDNodePtr node, float **query, vector<pointPtr>& res); // recursive descend

//	void boostWithPublicData(KDNodePtr node, MinPerTreePtr pub);
	void boostWithPublicData(KDNodePtr node, KDTreePtr pub);
	void boostWithPublicData(KDTreePtr pub);


	void calculateLeafTuples(); // calculates the aggregate leafTuples at each node
	void recCalculateLeafTuples(KDNodePtr node);

	void calculatePerceivedMicroTuples(); // calculates the perceived MT tuples at the leaves
	void recCalculatePerceivedMicroTuples(KDNodePtr node, double ratio);


	void countQuery(float **query, double& estimate, double& actual); // performs count range query
	void recCountQuery(KDNodePtr node, float **query, double& estimate, double& actual);

	void countQueryREF(float **query, double& estimate, double& actual); // performs count range query for REFINEMENT
	void recCountQueryREF(KDNodePtr node, float **query, double& estimate, double& actual);

	void countPerceivedMicroTuples();
	void recCountPerceivedMicroTuples(KDNodePtr node, double& count);


	friend ostream& operator<<(ostream& os, const KDTree*& t);
	friend ostream& operator<<(ostream& os, const KDTree& t);
//	friend bool overlaps(float **rect1, float **rect2);
//	friend bool inside(float* point, float ** rect);

private:
	bool groupsCalculated;
};

ostream& operator<<(ostream& os, const KDTreePtr& t);
ostream& operator<<(ostream& os, const KDTree& t);

string getString(float* point, int numDimensions);
bool overlaps(float **rect1, float **rect2, int numDimensions);
bool inside(float* point, float ** rect, int numDimensions);
// strict version:
bool overlapsStr(float **rect1, float **rect2, int numDimensions);
bool insideStr(float* point, float ** rect, int numDimensions);


#endif /*KDTREE_H_*/
