#include "minpertree.h"


using namespace std;

MinPerTree::MinPerTree(char *inputfile, int K){
	groupsCalculated = false;
	float *p;
	vector<float *> v;
	cerr << "MinPerTree:: reading from " << inputfile << endl;
	fstream fin(inputfile, ios::in);

	fin >> numDimensions; // first read dimensionality
	while(!fin.eof()){
		p = new float[numDimensions];
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		v.push_back(p);
	}
	fin.close();
	cerr << "... done reading " << v.size() << " points" << endl;

	this->K = K;

//	cout << "K=" << this->K << endl;

	int numPoints = v.size();
	pointPtr *points = new pointPtr[numPoints];
	for (int i=0; i<numPoints; i++){
		points[i] = new point(numDimensions, v[i], true);
	}
	v.clear();

	root = new MinPerNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "MinPerTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;

}


MinPerTree::MinPerTree(vector<pointPtr> poiset, int K){
	groupsCalculated = false;
	numDimensions = poiset[0]->numDimensions;
	this->K = K;

	int numPoints = poiset.size();
	pointPtr *points = new pointPtr[numPoints];
	copy(poiset.begin(), poiset.end(), points);

	root = new MinPerNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "MinPerTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;
}


void MinPerTree::recursiveSplit(MinPerNodePtr node){
//	cerr << "recursiveSplit @ " << node->id << endl;
	if (node->numPoints>=2*K){
		MinPerNodePtr left, right;
//		bool hasSplit = node->splitNodeNEW(left, right, sizeWorld, K);
		bool hasSplit = node->splitNode(left, right, sizeWorld, K);
////		bool hasSplit = node->splitNodeOLD(left, right, sizeWorld, K);

		if (hasSplit){
			nodes[left->id] = left;
			nodes[right->id] = right;
			left->calculateNumMicroPoints();
			right->calculateNumMicroPoints();

			if (left->numMicroPoints==0){
				nodes.erase(left->id);
				delete left;
				node->left = NULL;
			}
			else{
				recursiveSplit(left);
			}

			if (right->numMicroPoints==0){
				nodes.erase(right->id);
				delete right;
				node->right = NULL;
			}
			else{
				recursiveSplit(right);
			}

		}
	}
}


void MinPerTree::calculateSizeWorld(){
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}
}


vector<group> MinPerTree::getGroups(){
	if (!groupsCalculated){
		groups.clear();
		groupsCalculated = true;

		map<int, MinPerNodePtr>::iterator iter;
		for (iter=nodes.begin(); iter!=nodes.end(); iter++){
			if (!iter->second->leafNode){
				continue; // this is not a leaf node
			}
			group* gp = new group(numDimensions);

			for (int i=0; i<iter->second->numPoints; i++){
				gp->insertPoint(*(iter->second->points[i]));
			}
			groups.push_back(*gp);
		}
	}
	return groups;
}

void MinPerTree::writeGroups(char *inputfile){
	ostringstream kdfied;
	kdfied << "K" << K << "_" << inputfile;
	string writefile = kdfied.str();
	ofstream fout(writefile.c_str(), ios::out);

	map<int, MinPerNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		fout << iter->second << endl;
	}

	fout.close();
}


void MinPerTree::printMetrics(){
	map<int, MinPerNodePtr>::iterator iter;

	float normalPerimeter = 0;
	float sizeTimesPerimeter = 0;
	float discernability = 0;
	int numGroups = 0;

	float sizeWorld[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}

	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		numGroups++;
		discernability += iter->second->numMicroPoints*iter->second->numMicroPoints;

		float thisPerimeter = 0;
		for (int i=0; i<numDimensions; i++){
//			cout << "... " << (iter->second->extent[i][1] - iter->second->extent[i][0])/sizeWorld[i] << endl;
			float diff = iter->second->extent[i][1] - iter->second->extent[i][0];
			thisPerimeter += diff/sizeWorld[i];
		}
		normalPerimeter += thisPerimeter;
		sizeTimesPerimeter += thisPerimeter*iter->second->numMicroPoints;
	}

	cout << "there are " << numGroups << " groups" << endl;
	cout << "discernability = " << discernability << endl;
//	cout << "normalized average equivalence class size = " << ((float) K*nodes[1]->numMicroPoints)/numGroups << endl;
	cout << "total perimeter = " << normalPerimeter << endl;
	cout << "total (perimeter)*(numMicroPoint) = " << sizeTimesPerimeter << endl;

}


void MinPerTree::boostWithPublicData(MinPerNodePtr node, KDTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (!node->leafNode){
		boostWithPublicData(node->left, pub);
		boostWithPublicData(node->right, pub);
	}
	else{ // do the range query
//		cout << "do the range query" << endl;
		vector<pointPtr> newPoints = pub->rangeQuery(node->mbr);
//		cout << "query ok" << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
			(*resIter)->inMicro = false;
		}

		vector<pointPtr> thesePoints(node->numPoints);
		copy(node->points, node->points+node->numPoints, thesePoints.begin());

//		cout << "newPoints.size()=" << newPoints.size() << endl;
//		cout << "thesePoints.size()=" << thesePoints.size() << endl;

		vector<pointPtr> joinedPoints = SortMerge(newPoints, thesePoints);

//		cout << "joinedPoints.size()=" << joinedPoints.size() << endl;

		delete[] node->points;

		node->numPoints = joinedPoints.size();
		node->points = new pointPtr[node->numPoints];
		copy(joinedPoints.begin(), joinedPoints.end(), node->points);
		node->calculateNumMicroPoints();
//		cout << "numMicroPoints=" << node->numMicroPoints << endl;
		recursiveSplit(node);

	}
}


void MinPerTree::boostWithPublicData(MinPerNodePtr node, MinPerTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (!node->leafNode){
		boostWithPublicData(node->left, pub);
		boostWithPublicData(node->right, pub);
	}
	else{ // do the range query
//		cout << "do the range query" << endl;
		vector<pointPtr> newPoints = pub->rangeQueryStr(node->mbr);
//		cout << "query ok" << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
			(*resIter)->inMicro = false;
		}

		vector<pointPtr> thesePoints(node->numPoints);
		copy(node->points, node->points+node->numPoints, thesePoints.begin());

//		cout << "newPoints.size()=" << newPoints.size() << endl;
//		cout << "thesePoints.size()=" << thesePoints.size() << endl;

		vector<pointPtr> joinedPoints = SortMerge(newPoints, thesePoints);

//		cout << "joinedPoints.size()=" << joinedPoints.size() << endl;

		delete[] node->points;

		node->numPoints = joinedPoints.size();
		node->points = new pointPtr[node->numPoints];
		copy(joinedPoints.begin(), joinedPoints.end(), node->points);
		node->calculateNumMicroPoints();
//		cout << "numMicroPoints=" << node->numMicroPoints << endl;
		recursiveSplit(node);

	}
}


void MinPerTree::boostWithPublicData(MinPerTreePtr pub){
	groupsCalculated = false;
	map<int, MinPerNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (iter->second->leafNode){
			iter->second->boostedLeafNode = true;
		}
	}

//	int lStart = -1;
//	int lEnd = -1;
////	map<int, MinPerNodePtr>::iterator iter;
//	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
//		if (!iter->second->leafNode){
//			continue; // this is not a leaf node
//		}
//		if (lStart==-1) lStart = iter->first;
//		lEnd = iter->first;
//	}


	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->boostedLeafNode){
			continue; // this is not a leaf node that is going to be boosted
		}

		vector<pointPtr> result = pub->rangeQuery(iter->second->mbr);
//		cout << "newPoints.size()=" << result.size() << endl;
//		cout << "thesePoints.size()=" << iter->second->numPoints << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=result.begin(); resIter!=result.end(); resIter++){
			(*resIter)->inMicro = false;
		}
//		cout << "range:: | ";
//		for (int j=0; j<numDimensions; j++){
//			cout << iter->second->extent[j][0] << " -- " << iter->second->extent[j][1] << " | ";
//		}
//		cout << endl;
//		cout << "for node " << iter->second->id << " rangeQueryStr returned " << result.size() << " results" << endl;
//		for (uint i=0; i<result.size(); i++){
//			cout << *(result[i]) << " ";
//		}
//		cout << endl;

		for (int i=0; i<iter->second->numPoints; i++){
			pointPtr mtp = iter->second->points[i];
			for (resIter=result.begin(); resIter!=result.end(); resIter++){
				if ( *mtp == *(*resIter) ){
					(*resIter)->inMicro = true;
					break;
				}
			}
		}

		pointPtr *newpoints = new pointPtr[result.size()];
//		cout << "joinedPoints.size()=" << result.size() << endl;
		int tc = 0;
		for (uint i=0; i<result.size(); i++){
			newpoints[i] = new point(*(result[i]));
			if (newpoints[i]->inMicro) tc++;
		}
//		cout << "tc=" << tc << endl;

		for (int i=0; i<iter->second->numPoints; i++){
			delete iter->second->points[i];
		}
		delete[] iter->second->points;

		iter->second->points = newpoints;
		iter->second->numMicroPoints = iter->second->numPoints;
		iter->second->numPoints = result.size();
		iter->second->calculateNumMicroPoints();
//		cout << "numMicroPoints=" << iter->second->numMicroPoints << endl;

		recursiveSplit(iter->second);

	}

}


vector<pointPtr> MinPerTree::rangeQuery(float **query){
	vector<pointPtr> result;
//	cout << "range: " << getString(query, numDimensions) << endl;
	recRangeQuery(root, query, result);

	return result;
}

void MinPerTree::recRangeQuery(MinPerNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << endl;
//	if ( node->left==NULL && node->right==NULL ){ // leaf node
	if ( node->leafNode ){
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
			if ( inside(node->points[i]->p, query, numDimensions) ){ // point inside
				res.push_back(node->points[i]);
//				cout << "in " << *(node->points[i]) << endl;
			}
			else{
//				cout << "out " << *(node->points[i]) << endl;
			}
		}
	}
//	cout << "gonna descend!" << endl;
	if ( node->left!=NULL && overlaps(query, node->left->extent, numDimensions) ){
		recRangeQuery(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlaps(query, node->right->extent, numDimensions) ){
		recRangeQuery(node->right, query, res); // descend
	}

}

vector<pointPtr> MinPerTree::rangeQueryStr(float **query){
	vector<pointPtr> result;

	recRangeQueryStr(root, query, result);

	return result;
}

void MinPerTree::recRangeQueryStr(MinPerNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << endl;
	if ( node->left==NULL && node->right==NULL ){ // leaf node
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
			if ( insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				res.push_back(node->points[i]);
//				cout << "in " << getString(node->points[i], numDimensions) << endl;
			}
			else{
//				cout << "out " << getString(node->points[i], numDimensions) << endl;
			}
		}
	}
	if ( node->left!=NULL && overlapsStr(query, node->left->extent, numDimensions) ){
		recRangeQueryStr(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlapsStr(query, node->right->extent, numDimensions) ){
		recRangeQueryStr(node->right, query, res); // descend
	}

}





ostream& operator<<(ostream& os, const MinPerTreePtr& t){
	map<int, MinPerNodePtr>::iterator iter;

	cout << endl;
	for(iter = t->nodes.begin(); iter!=t->nodes.end(); iter++){
        cout << "map id=" << iter->first << " " << iter->second << endl;
    }

	return os;
}

