#ifndef MinPerTree_H_
#define MinPerTree_H_

#include <fstream>
#include <sstream>
#include <map>
#include <set>
#include <string>
#include <algorithm>
#include "minpernode.h"
#include "../kdtree/kdtree.h"
#include "../common/group.h"
#include "../common/point.h"
#include "../common/Timing.h"

using namespace std;


typedef class MinPerTree *MinPerTreePtr;

class MinPerTree
{
public:
	int numDimensions; // dimensionality
	
	map<int, MinPerNodePtr> nodes;
//	vector<NodePtr> nodes;
	
	int K; // anonymity requirement
	
	vector<group> groups; // K-anonymized groups
	
	ofstream fout;
	MinPerNodePtr root;
	
	float* sizeWorld;
	
	MinPerTree(char *inputfile, int K);
	MinPerTree(vector<pointPtr> poiset, int K);
		
	void recursiveSplit(MinPerNodePtr node);
	
	
	void calculateSizeWorld();
	
	void writeGroups(char *inputfile);
	vector<group> getGroups();
	void printMetrics();
	
	vector<pointPtr> rangeQuery(float **query);
	void recRangeQuery(MinPerNodePtr node, float **query, vector<pointPtr>& res); // recursive descend
	
	// strict versions:
	vector<pointPtr> rangeQueryStr(float **query);
	void recRangeQueryStr(MinPerNodePtr node, float **query, vector<pointPtr>& res); // recursive descend
	
	void boostWithPublicData(MinPerNodePtr node, MinPerTreePtr pub);
	void boostWithPublicData(MinPerNodePtr node, KDTreePtr pub);
	void boostWithPublicData(MinPerTreePtr pub);
	
	friend ostream& operator<<(ostream& os, const MinPerTree*& t);
	friend ostream& operator<<(ostream& os, const MinPerTree& t);
//	friend bool overlaps(float **rect1, float **rect2);
//	friend bool inside(float* point, float ** rect);

private:
	bool groupsCalculated;
};

ostream& operator<<(ostream& os, const MinPerTreePtr& t);
ostream& operator<<(ostream& os, const MinPerTree& t);

string getString(float* point, int numDimensions);
bool overlaps(float **rect1, float **rect2, int numDimensions);
bool inside(float* point, float ** rect, int numDimensions);
// strict version:
bool overlapsStr(float **rect1, float **rect2, int numDimensions);
bool insideStr(float* point, float ** rect, int numDimensions);


#endif /*MinPerTree_H_*/
