#include "kdtree/kdtree.h"


using namespace std;

int main(int argc, char** argv){

	if (argc!=4) {
		cerr << "Usage: " << argv[0] << " [microdata] [K] [selectivity]" << endl;
		cerr << "K is anonymity requirement" << endl;
		exit(-1);
    }
	int K = atoi(argv[2]);
	int numDimensions;
	double selectivity = atof(argv[3]);
	vector<pointPtr> points = readFromFile(argv[1], true, numDimensions);
	Timing* tim = new Timing();

	cout << "Mondrian" << endl;
	KDTreePtr tree;
	int iterations = 1;
	tim->startClock();
	for (int i=0; i<iterations; i++){
		tree = new KDTree(points, K);
	}
	long elapsed = tim->stopClock();

	tree->printMetrics();


	vector<group> groups = tree->getGroups();

	calculateMetrics(groups, tree->sizeWorld);





//	cout << "performing queries..." << endl;
	double sumError = 0, sumActual = 0, maxActual = 0, sumRelError = 0, maxRelError = 0;


	ofstream mseFile, relErrFile;
	mseFile.open("MSE.out", ios::out | ios::app);
	relErrFile.open("RelErr.out", ios::out | ios::app);


//	srand(456);


	// load points into a kdtree for getting correct answers
	KDTreePtr cor = new KDTree(argv[1], 2);

	int numQueries = 0;
	while (numQueries<100){
		double estimate=0, actual=0, estimateCOR=0, actualCOR=0;
		float **query;
		query = new float*[numDimensions];
		for (int d=0; d<numDimensions; d++){
			double rnd = ((double) rand())/RAND_MAX;
			query[d] = new float[2];
			double dimSize = tree->sizeWorld[d];
//			if (dimSize>1000000) dimSize = 1000000; // fix for IPUMS dataset
			query[d][0] = tree->nodes[1]->extent[d][0] + rnd * (1-selectivity) * dimSize;
			query[d][1] = query[d][0] + selectivity * dimSize;
//			cout << "[" << query[d][0] << " : " << query[d][1] << "] ";
		}
//		cout << endl << "query " << i << endl;
		tree->countQuery(query, estimate, actual);
		cor->countQuery(query, estimateCOR, actualCOR);
		if (actual < 1) continue; //skip this query
		cout << "... estimate=" << estimate << " actual=" << actual << endl;
		cout << "*** actualCOR=" << actualCOR << endl << endl;
		sumError += (actual-estimate)*(actual-estimate);
		sumActual += actual;
		sumRelError += fabs(actual-estimate)/actual;
		if (fabs(actual-estimate)/actual > maxRelError ){
			maxRelError = fabs(actual-estimate)/actual;
			maxActual = actual;
		}
		for (int d=0; d<numDimensions; d++){
//			cout << tree->sizeWorld[d] << " ";
			delete query[d];
		}
//		cout << endl;
		delete query;
		numQueries++;
	}
	cout << "AVG SQUARED ERROR " << sumError/numQueries << endl;
	mseFile << sumError/numQueries << "\t";
	cout << "AVG RELATIVE ERROR " << sumRelError/numQueries << endl;
	relErrFile << sumRelError/numQueries << "\t";
	cout << "MAX RELATIVE ERROR " << maxRelError << " at " << maxActual << endl;

	cout << "Elapsed Time: " << ((double)elapsed)/iterations << endl;
	cout << "------------------------------------" << endl;
	cout << endl;

	mseFile.close();
	relErrFile.close();

}
