#include "stripeNode.h"




stripeNode::stripeNode(vector<int> remainingDims,  vector<pointPtr> points){
	this->remainingDims = remainingDims;
	this->points = points;
	calculateNumMicroPoints();
	leafNode = true;
	parent = NULL;
//	id = stripeNode::next_id++;
//	cout << "node with id=" << id << " created" << endl;
}



int stripeNode::calculateNumMicroPoints(){
	numMicroPoints = 0;
	for (uint i=0; i<points.size(); i++){
		if (points[i]->inMicro){
			numMicroPoints++;
		}
	}
	return numMicroPoints;
}


int stripeNode::pickDimension(int q){
//	if (remainingDims.size()==1) return remainingDims[0];
	if (remainingDims.size()>0) return remainingDims[0];
	
//	cout << "-------------------" << endl;
//	cout << "node id " << id << endl;
//	cout << "dims to examine " << remainingDims.size() << endl;
	vector<int>::iterator iter;
	
	int bestDim = -1;
	float bestVariance = FLT_MAX;
	
	for (iter=remainingDims.begin(); iter!=remainingDims.end(); iter++){
		
		int splitDim = *iter;
		
		orderByDim *myorder = new orderByDim(splitDim);
		sort(points.begin(), points.end(), (*myorder));
		delete myorder;
	
		int depth = (int) ((float)points.size())/q;
		
		
		
		int count = 0;
		bool readyToSplit = false;
		float someStripeLength;
		float prevSplitValue = points[0]->p[splitDim];
		
		float sum = 0;
		float sumsq = 0;
		int numQuantiles = 0;
		
		
		for (uint i=0; i<points.size(); i++){
			count++;
	//		cout << "i=" << i << " count=" << count << endl;
			if (readyToSplit){
				if (points[i]->p[splitDim]==points[i-1]->p[splitDim]){
					continue;
				}
				else{
					readyToSplit = false;
					someStripeLength = points[i]->p[splitDim] - prevSplitValue;
					sum += someStripeLength;
					sumsq += someStripeLength*someStripeLength;
					prevSplitValue = points[i]->p[splitDim];
					numQuantiles++;
					count = 1;
					continue;
				}
			}
			else{
				if (count==depth){
					readyToSplit = true;
				}	
			}
					
		}
	//	cout << "last stripe has " << someStripe.size() << " points" << endl;
		
		if (readyToSplit){
			someStripeLength = points[points.size()-1]->p[splitDim] - prevSplitValue;
			sum += someStripeLength;
			sumsq += someStripeLength*someStripeLength;
			numQuantiles++;
		}
		else{
			sum -= someStripeLength;
			sumsq -= someStripeLength*someStripeLength;
			
			someStripeLength = points[points.size()-1]->p[splitDim] - (prevSplitValue - someStripeLength);
			sum += someStripeLength;
			sumsq += someStripeLength*someStripeLength;
		}
		
		
		float thisVariance = (sumsq - sum*sum/numQuantiles)/(numQuantiles-1);
		
//		float thisVariance = q - numQuantiles;
		
//		cout << "thisVariance=" << thisVariance << endl;
		
		if (thisVariance<bestVariance){
			bestVariance = thisVariance;
			bestDim = splitDim;
		}
		
	}
	
	return bestDim;
	
//	return remainingDims[0];	
} 


bool stripeNode::splitNode(int q, int K, int numDimensions, float* sizeWorld){
//	cout << "splitNode called" << endl;

	
//	leafNode = false;
//	if (remainingDims.size()==1) return finalSplitNode(q, K, numDimensions, sizeWorld);
	
	
	splitDim = pickDimension(q);
	vector<int> dims = remainingDims;
	
	vector<vector<pointPtr> > theStripes; // to become children
	
	orderByDim *myorder = new orderByDim(splitDim);
	sort(points.begin(), points.end(), (*myorder));
	
	delete myorder;
	
	int depth = (int) ((float)points.size())/q;
	if (dims.size()==1){ // last split, create groups of size K
		depth = K; 
	}
//	cout << "q=" << q << " depth=" << depth << endl;
	
	int count = 0;
	bool readyToSplit = false;
	vector<pointPtr> someStripe;
	
	for (uint i=0; i<points.size(); i++){
		count++;
//		cout << "i=" << i << " count=" << count << endl;
		if (readyToSplit){
//			cout << "someStripe.size()=" << someStripe.size() << endl; 
			if (points[i]->p[splitDim]==someStripe[someStripe.size()-1]->p[splitDim]){
				someStripe.push_back(points[i]);
				continue;
			}
			else{
				readyToSplit = false;
				theStripes.push_back(someStripe);
//				cout << "new Stripe with size=" << someStripe.size() << endl; 
				someStripe.clear();
				someStripe.push_back(points[i]);
				count = 1;
				continue;
			}
		}
		else{
			someStripe.push_back(points[i]);
			if (count==depth){
				readyToSplit = true;
			}	
		}
				
	}
//	cout << "last stripe has " << someStripe.size() << " points" << endl;
	
	if (someStripe.size()<(uint)depth){
		if (theStripes.size()>0){
			for (uint i=0; i<someStripe.size(); i++){
				theStripes[theStripes.size()-1].push_back(someStripe[i]);
			}
//			cout << "previous Stripe adjusted to size=" << theStripes[theStripes.size()-1].size() << endl;
		}
		else{
//			cerr << "numDimsRemaining = " << remainingDims.size() << " we're in shit!" << endl;
			return false;
		}
	}
	else{
//		cout << "final Stripe with size=" << someStripe.size() << endl;
		theStripes.push_back(someStripe);
	}
	
//	cout << "created " << theStripes.size() << " stripes" << endl;
	
//	for (uint i=0; i<theStripes.size(); i++){
//		cout << "stripe " << i << " has " << theStripes[i].size() << " entries" << endl;
//		for (uint j=0; j<theStripes[i].size(); j++){
//			cout << *(theStripes[i][j]) << " ";
//		}
//		cout << endl;
//	}
	
	
//	cout << "this node has left dims " << dims.size() << endl;
	// remove dimension from set of dimensions
	vector<int>::iterator iter;
	for (iter=dims.begin(); iter!=dims.end(); iter++){
		if (*iter==splitDim){
			dims.erase(iter);
			break;
		}
	}
//	cout << "after, this node has left dims " << dims.size() << endl;
	
	for (uint i=0; i<theStripes.size(); i++){
		stripeNodePtr newnode = new stripeNode(dims, theStripes[i]);
		newnode->parent = this;
		children.push_back(newnode);
	}
	
	return true;
}


bool stripeNode::finalSplitNode(int q, int K, int numDimensions, float* sizeWorld){
//	cout << "this Node has " << points.size() << " points" << endl; 
	
	float bestScore;
	
	int thisD = remainingDims[0];
	
	orderByDim *myorder = new orderByDim(thisD);
	sort(points.begin(), points.end(), (*myorder));
	delete myorder;
	
	list<pointPtr> leftie;
	list<pointPtr> rightie;
	int leftieSize = 0;
	int rightieSize = 0;
	int leftieMicroPoints = 0;
	int rightieMicroPoints = 0;
	
	float** leftExtent;
	float** rightExtent;
	

	int bestLeftNumPoints;
	int start = 0;
	int curr;
	
	bool done = false;
	
	while(!done){
		bestScore = FLT_MAX;
		bestLeftNumPoints = 0;
		curr = start;
//		cout << "curr=" << curr << endl;
	
		float firstValue = points[start]->p[thisD];
	
		bool firstTime = true;
		
		leftie.clear();
		leftieSize = 0;
		
		rightie.clear();
		rightieSize = 0;
		for (uint i=start; i<points.size(); i++){
			rightie.push_back(points[i]);
			rightieSize++;
			if (points[i]->inMicro) rightieMicroPoints++;
		}
		rightExtent = findExtent(rightie);
		
	
		while ((uint)curr<points.size()){
//			cout << "curr=" << curr << " start=" << start << endl;
			
//			cerr << "#"; fflush(stderr);

			float splitValue = points[curr]->p[thisD];

			bool calcRightExtent = false;
		
			float nextSplitValue = points[points.size()-1]->p[thisD];
			for (uint tmp=curr; tmp<points.size(); tmp++){
				if (points[tmp]->p[thisD]!=splitValue){
					nextSplitValue = points[tmp]->p[thisD];
					break;
				}
//				if (tmp==points.size()-1){
//					nextSplitValue = points[points.size()-1]->p[thisD];
//				}
			}
			rightExtent[thisD][0] = nextSplitValue;
		
		
			while (points[curr]->p[thisD] == splitValue){
//				cerr << "-"; fflush(stderr);
				if (points[curr]->p[thisD] != firstValue){
					float** temp = newExtent(leftExtent, points[curr]);
					for (int d=0; d<numDimensions; d++){
						delete[] leftExtent[d];
					}
					delete[] leftExtent;
					leftExtent = temp;
				}
				leftie.push_back(points[curr]);
				leftieSize++;
				if (points[curr]->inMicro) leftieMicroPoints++;
	//				cout << "*(iter->second):" << *(iter->second) << " *(rightie.front()):" << *(rightie.front()) << endl;
				assert( (*(points[curr])==*(rightie.front())) );
				rightie.pop_front();
				rightieSize--;
				if (points[curr]->inMicro) rightieMicroPoints--;
				if (rightieSize<K){
	//					cerr << "(R<K)"; fflush(stderr);
					break;
				}
				if (!calcRightExtent && !insideStr(points[curr]->p, rightExtent, numDimensions) ){
					calcRightExtent = true;
	//					cerr << "*"; fflush(stderr);
				}
				curr++;		
			}
			
			if (firstTime){
				firstTime = false;
				leftExtent = findExtent(leftie);
			}
			
			if (calcRightExtent){
				for (int d=0; d<numDimensions; d++){
					delete[] rightExtent[d];
				}
				delete[] rightExtent;
				rightExtent = findExtent(rightie);
			}
			
	//			cout << "leftExtent:" << getString(leftExtent, numDimensions) << " rightExtent:" << getString(rightExtent, numDimensions) << endl;
			
			if (leftieSize<K){
//					cerr << "(L<K)"; fflush(stderr);
				continue;
			}
			if (rightieSize<K){
//					cerr << "(R<K)"; fflush(stderr);
				break;
			}
			
			float score = getNCPFast(leftExtent, sizeWorld, numDimensions, leftieMicroPoints) + getNCPFast(rightExtent, sizeWorld, numDimensions, rightieMicroPoints);
			
			if (score<bestScore){
				bestScore = score;
				bestLeftNumPoints = curr-start;
			}
			
		}
		
		for (int d=0; d<numDimensions; d++){
			delete[] leftExtent[d];
		}
		delete[] leftExtent;
		
		
		vector<pointPtr> leftPoints;

		
		
//		cout << "bestLeftNumPoints=" << bestLeftNumPoints << endl;
		
		
		if (bestLeftNumPoints>0){
			
			
//			cout << "left:: " << start << "-" << start+bestLeftNumPoints-1 << endl;
			
			for (uint i=start; i<points.size(); i++){
		//		cerr << i << " "; fflush(stderr);
				if (i < start+bestLeftNumPoints){
					leftPoints.push_back(points[i]);
				}
				else{
					break;
				}
			}
		
		
			vector<int> dummyDims;
			stripeNodePtr newchild = new stripeNode(dummyDims, leftPoints);
			newchild->parent = this;
			children.push_back(newchild);
		}	
				
			if (points.size()-(start+bestLeftNumPoints)<2*K || bestLeftNumPoints==0){
				done = true;
				
				vector<pointPtr> rightPoints;
				
				
//				cout << "right:: " << start+bestLeftNumPoints << "-" << points.size()-1 << endl;
				
				for (uint i=start+bestLeftNumPoints; i<points.size(); i++){
			//		cerr << i << " "; fflush(stderr);
					rightPoints.push_back(points[i]);
				}
				
				vector<int> dummyDims;
				stripeNodePtr newchild = new stripeNode(dummyDims, rightPoints);
				newchild->parent = this;
				children.push_back(newchild);
				
			}
			
			start += bestLeftNumPoints;
//			cout << "start becomes " << start << endl;
		
//		}
//		else{
//			done = true;
//		}
		
	}	
	
	
	for (int d=0; d<numDimensions; d++){
		delete[] rightExtent[d];
	}
	delete[] rightExtent;
	
	
//	cout << "exited finalSplitNode" << endl;
	return true;
}

float stripeNode::getNCPFast(float** extent, float* sizeWorld, int numDimensions, int numMicroInCluster){
//	cerr << "."; fflush(stderr);
	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
	}
	
	return numMicroInCluster;
//	return perimeter*numMicroInCluster;
//	return perimeter;
}



