#include "stripes.h"




Stripes::Stripes(char *inputfile, int K){
	groupsCalculated = false;
	this->K = K;
	
	cerr << "Stripes:: reading from " << inputfile << endl;
	fstream fin(inputfile, ios::in);
	
	fin >> numDimensions; // first read dimensionality
	while(!fin.eof()){
		float *p = new float[numDimensions];
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		pointPtr po = new point(numDimensions, p, true);
		delete p;
		points.push_back(po);
	}
	fin.close();
	cerr << "... done reading " << points.size() << " points" << endl;
	
	q = (int) floor(pow((((float)points.size())/K), (float) 1/numDimensions));
	cout << "previously q=" << q << endl;
	if (q<3) q=3;
	cerr << "N=" << points.size() << " K=" << K << " d=" << numDimensions << " q=" << q << endl; 
//	cout << "hi!" << endl;
	extent = getExtent(points);
//	cout << "ok" << endl;
	calculateSizeWorld();
	
	run();
}


Stripes::Stripes(vector<pointPtr> poiset, int K){
	groupsCalculated = false;
	this->K = K;
	numDimensions = poiset[0]->numDimensions;
	points = poiset;
	
	q = (int) floor(pow((((float)points.size())/K), (float) 1/numDimensions));
	cout << "previously q=" << q << endl;
	if (q<3) q=3;
	cerr << "N=" << points.size() << " K=" << K << " d=" << numDimensions << " q=" << q << endl; 
	
	extent = getExtent(points);
	calculateSizeWorld();
	
	run();
}


void Stripes::run(){
	vector<int> dims;
	for (int i=0; i<numDimensions; i++){
		dims.push_back(i);
	}
	
	
//	do{
		root = new stripeNode(dims, points);
		recursiveSplit(root);
//		vector<group> thegroups = getGroups();
//		calculateMetrics(thegroups, sizeWorld);
//		thegroups.clear();
//	} while(next_permutation(dims.begin(), dims.end()));
	

}


void Stripes::boostWithPublicData(stripeNodePtr node, MinPerTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (node->remainingDims.size()>1){
//		cout << "children.size()=" << node->children.size() << endl;
		for (uint i=0; i<node->children.size(); i++){
			boostWithPublicData(node->children[i], pub);
		}
	}
	else{ // do the range query
//		cout << "children.size()=" << node->children.size() << endl;	
		node->children.clear();
		
		float** mbb = getExtent(node->points);
		vector<pointPtr> newPoints = pub->rangeQuery(mbb);
//		cout << "query returned" << endl;
		for (int d=0; d<numDimensions; d++){
			delete[] mbb[d];
		}
		delete[] mbb;
		
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
			(*resIter)->inMicro = false;
		}
		
		
		vector<pointPtr> resultPoints(newPoints);
		
		vector<pointPtr> thesePoints(node->points);
		
		vector<pointPtr> joinedPoints = SortMerge(resultPoints, thesePoints);
		
		node->points.clear();
		node->points = joinedPoints;
		
//		cout << "do the recursiveSplit" << endl;
		recursiveSplit(node);
//		cout << "recursiveSplit OK" << endl;
	}
	
	
}


void Stripes::boostWithPublicData(stripeNodePtr node, KDTreePtr pub){
//	cout << "boostWithPublicData" << endl;
//	if (node->remainingDims.size()>1){
	if (!node->leafNode){
//		cout << "children.size()=" << node->children.size() << endl;
		for (uint i=0; i<node->children.size(); i++){
			boostWithPublicData(node->children[i], pub);
		}
	}
	else{ // do the range query
//		cout << "children.size()=" << node->children.size() << endl;	
		node->children.clear();
		
		float** mbb = getExtent(node->points);
		vector<pointPtr> newPoints = pub->rangeQuery(mbb);
//		cout << "query returned" << endl;
		for (int d=0; d<numDimensions; d++){
			delete[] mbb[d];
		}
		delete[] mbb;
		
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
			(*resIter)->inMicro = false;
		}
		
//		cout << "there are " << newPoints.size() << " points in total" << endl;
		
		vector<pointPtr> resultPoints(newPoints);
		
		vector<pointPtr> thesePoints(node->points);
		
		vector<pointPtr> joinedPoints = SortMerge(resultPoints, thesePoints);
//		cout << "there are " << joinedPoints.size() << " points in join" << endl;
		
		node->points.clear();
		node->points = joinedPoints;
		
		vector<int> newDims;
		for (int i=0; i<numDimensions; i++){
			newDims.push_back(i);
		}
		
//		newDims = node->parent->remainingDims;
		
		node->remainingDims = newDims;
		
		// calculate new q
		
		int localQ = (int) floor(pow((((float)joinedPoints.size())/K), (float) 1/numDimensions));
//		cout << "previously localQ=" << localQ << endl;
		if (localQ<3) localQ=3;
//		cerr << "localN=" << joinedPoints.size() << " localQ=" << localQ << endl; 
	
		
		
//		cout << "do the recursiveSplit" << endl;
		recursiveSplit(node, localQ, K);
//		cout << "recursiveSplit OK" << endl;
	}
	
	
}



Stripes::~Stripes(){
	for (int d=0; d<numDimensions; d++){
		delete[] extent[d];
	}
	delete[] extent;
}



float** Stripes::getExtent(vector<pointPtr> poiset){
	float** extent = new float* [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
		for (uint i=0; i<poiset.size(); i++){
			temp.push_back(poiset[i]->p[d]);
		}		
		extent[d][0] = *min_element(temp.begin(), temp.end());
		extent[d][1] = *max_element(temp.begin(), temp.end());
		temp.clear();
	}
	return extent;
}

/*
int Stripes::pickDimension(vector<pointPtr> poiset, vector<int> dims){
	return dims[0];
//	int bestD = -1;
//	for (uint i=0; i<dims.size(); i++){
//		int D = dims[i];
//		int* freqs = new int[]
//		
//	}
}
*/

/*
void Stripes::recurseStrip(vector<pointPtr> poiset, vector<int> dims){
	if (dims.size()==0){
		return createGroup(poiset);
//		 return; // stop recurse
	}
	
	// pick dimension to strip!
	int D = pickDimension(poiset, dims);
//	cout << "D=" << D << endl;
	
	vector<vector<pointPtr> > theStripes;
	
	orderByDim *myorder = new orderByDim(D);	
	sort(poiset.begin(), poiset.end(), (*myorder));
	
	delete myorder;
	
	int depth = (int) ((float)poiset.size())/q;
	if (dims.size()==1){ // last split
		depth = K; 
	}
//	cout << "q=" << q << " depth=" << depth << endl;
	
	int count = 0;
	bool readyToSplit = false;
	vector<pointPtr> someStripe;
	
	for (uint i=0; i<poiset.size(); i++){
		count++;
//		cout << "i=" << i << " count=" << count << endl;
		if (readyToSplit){
			if (poiset[i]->p[D]==someStripe[someStripe.size()-1]->p[D]){
				someStripe.push_back(poiset[i]);
				continue;
			}
			else{
				readyToSplit = false;
				theStripes.push_back(someStripe);
				someStripe.clear();
				someStripe.push_back(poiset[i]);
				count = 1;
				continue;
			}
		}
		else{
			someStripe.push_back(poiset[i]);
			if (count==depth){
				readyToSplit = true;
			}	
		}
				
	}
//	cout << "last stripe has " << someStripe.size() << " points" << endl;
	
	if (someStripe.size()<(uint)depth){
		if (theStripes.size()>0){
			for (uint i=0; i<someStripe.size(); i++){
				theStripes[theStripes.size()-1].push_back(someStripe[i]);
			}
		}
	}
	else{
		theStripes.push_back(someStripe);
	}
	
//	cout << "created " << theStripes.size() << " stripes" << endl;
	
	for (uint i=0; i<theStripes.size(); i++){
//		cout << "stripe " << i << " has " << theStripes[i].size() << " entries" << endl;
//		for (uint j=0; j<theStripes[i].size(); j++){
//			cout << *(theStripes[i][j]) << " ";
//		}
//		cout << endl;

	}
	
	vector<int>::iterator iter;
	for (iter=dims.begin(); iter!=dims.end(); iter++){
		if (*iter==D){
			dims.erase(iter);
			break;
		}
	}
	
	for (uint i=0; i<theStripes.size(); i++){
		recurseStrip(theStripes[i], dims);	
	}
	
}
*/

/*
void Stripes::createGroup(vector<pointPtr> poiset){
//	cout << "createGroup called" << endl;
	group* g = new group(numDimensions);
	
	for (uint i=0; i<poiset.size(); i++){
		g->insertPoint(*(poiset[i]));
	} 
	groups.push_back(*g);	
}
*/

void Stripes::calculateSizeWorld(){
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = extent[i][1] - extent[i][0];
	}
}


void Stripes::recursiveSplit(stripeNodePtr node){
//	cerr << "recursiveSplit @ " << node->id << endl;
	
	if (node->remainingDims.size()>0){
		bool hasSplit = node->splitNode(q, K, numDimensions, sizeWorld);
		
		if (hasSplit){
			node->leafNode = false;
			groupsCalculated = false;
//			cout << "there are " << node->children.size() << " children" << endl;
			for (uint i=0; i<node->children.size(); i++){
				recursiveSplit(node->children[i]);	
			}
		}
	}
	
}


void Stripes::recursiveSplit(stripeNodePtr node, int q, int K){
//	cerr << "recursiveSplit @ " << node->id << endl;
	
	if (node->remainingDims.size()>0){
		bool hasSplit = node->splitNode(q, K, numDimensions, sizeWorld);
//		cout << "splitNode returned hasSplit=" << hasSplit << endl;
		
		if (hasSplit){
			node->leafNode = false;
			groupsCalculated = false;
//			cout << "there are " << node->children.size() << " children" << endl;
			for (uint i=0; i<node->children.size(); i++){
				recursiveSplit(node->children[i], q, K);	
			}
		}
	}
	
}


vector<group> Stripes::getGroups(){
	if (!groupsCalculated){
		groups.clear();
		groupsCalculated = true;
		getGroupsRec(root);
	}
	return groups;
}

void Stripes::getGroupsRec(stripeNodePtr node){
	if (!node->leafNode){
		for (uint i=0; i<node->children.size(); i++){
			getGroupsRec(node->children[i]);	
		}
	}
	else{
//		cout << "group with " << node->points.size() << endl;
		group* gp = new group(numDimensions);
		for (uint i=0; i<node->points.size(); i++){
			gp->insertPoint(*(node->points[i]));
		}	
		groups.push_back(*gp);
	}
}


