#include "tdnode.h"


TDNode::TDNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension){
	this->numDimensions = numDimensions;
	this->numPoints = numPoints;
	this->points = points;
	this->splitDimension = splitDimension;
	this->numMicroPoints = calculateNumMicroPoints();
	splitValue = -2007;
	left = NULL;
	right = NULL;
	parent = NULL;
	leafNode = true;
	boostedLeafNode = false;
	isLeftChild = true;
	leafTuples = 0;
	perceivedMicroTuples = 0;
}

TDNode::~TDNode(){
	for (int i=0; i<numDimensions; i++){
		delete[] extent[i];
		delete[] mbr[i];
	}
	delete[] extent;
	delete[] mbr;
//	for (int i=0; i<numPoints; i++){
//		delete points[i];
//	}
	delete[] points;
}

int TDNode::calculateNumMicroPoints(){
	numMicroPoints = 0;
	for (int i=0; i<numPoints; i++){
		if (points[i]->inMicro){
			numMicroPoints++;
		}
	}
	return numMicroPoints;
}

void TDNode::calculateExtentMBR(float pad){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}

		vector<float>::iterator it = min_element(temp.begin(), temp.end());

		mbr[d][0] = *min_element(temp.begin(), temp.end()) - pad;
		mbr[d][1] = *max_element(temp.begin(), temp.end()) + pad;

		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}

		temp.clear();
	}
}


void TDNode::calculateExtentMBR(){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}

		vector<float>::iterator it = min_element(temp.begin(), temp.end());

		mbr[d][0] = *min_element(temp.begin(), temp.end());
		mbr[d][1] = *max_element(temp.begin(), temp.end());

		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}

		temp.clear();
	}
}


bool TDNode::splitNode(TDNodePtr &left, TDNodePtr &right, float* theSizeWorld, int K){
//	cerr << "splitNode (numPoints=" << numPoints << ") "; fflush(stderr);

	pointPtr seedA, seedB;

	seedA = points[0];

	for (int itr=0; itr<5; itr++){ // search for seeds
		float maxdist = -1;
		for (int i=0; i<numPoints; i++){
			if (!points[i]->inMicro) continue;
			float thisdist = 0;
			for (int d=0; d<numDimensions; d++){
				thisdist += fabs(seedA->p[d] - points[i]->p[d]);
			}
			if (thisdist>maxdist){
				maxdist = thisdist;
				seedB = points[i];
			}
		}

		maxdist = -1;
		for (int i=0; i<numPoints; i++){
			if (!points[i]->inMicro) continue;
			float thisdist = 0;
			for (int d=0; d<numDimensions; d++){
				thisdist += fabs(seedB->p[d] - points[i]->p[d]);
			}
			if (thisdist>maxdist){
				maxdist = thisdist;
				seedA = points[i];
			}
		}
	}

	if (*seedA==*seedB) return false; // no split if we got the same seeds

//	cout << "seedA " << *seedA << endl;
//	cout << "seedB " << *seedB << endl;


	vector<pointPtr> clusterA;
	vector<pointPtr> clusterB;
	clusterA.push_back(seedA);
	clusterB.push_back(seedB);



	float** extentA = findExtent(clusterA);
	float** extentB = findExtent(clusterB);


	float NCPclusterA = getNCP(clusterA, theSizeWorld);
	float NCPclusterB = getNCP(clusterB, theSizeWorld);

//	cout << "clusterA has " << clusterA.size() << " points" << endl;
//	cout << "clusterB has " << clusterB.size() << " points" << endl;
//
//	cout << "extentA: " << getString(extentA, numDimensions) << endl;
//	cout << "extentB: " << getString(extentB, numDimensions) << endl;

	int numMicroPointsA = 1;
	int numMicroPointsB = 1;


	bool lookedSeedA = false;
	bool lookedSeedB = false;

	for (int i=0; i<numPoints; i++){
//		if (i%10000==0 && i>0) cerr << i << " ";
//		cout << i << " " << endl; fflush(stdout);
		if (*(points[i])==*seedA && !lookedSeedA){
			lookedSeedA = true;
			continue;
		}
		if (*(points[i])==*seedB && !lookedSeedB){
			lookedSeedB = true;
			continue;
		}

//		cout << clusterA.size() << endl;
//		cout << clusterB.size() << endl;
//		cout << "cluster A before " << getNCP(clusterA, theSizeWorld) << endl;
//		cout << "cluster A after " << getNCP(clusterA, points[i], theSizeWorld) << endl;
//		cout << "cluster B before " << getNCP(clusterB, theSizeWorld) << endl;
//		cout << "cluster B after " << getNCP(clusterB, points[i], theSizeWorld) << endl;

		float** newExtentA;
		float** newExtentB;

		float incA = getNCPIncrease(extentA, newExtentA, NCPclusterA, points[i], numMicroPointsA, theSizeWorld);
		float incB = getNCPIncrease(extentB, newExtentB, NCPclusterB, points[i], numMicroPointsB, theSizeWorld);

//		cout << "ok" << endl;

		if (incA<incB){
			for (int d=0; d<numDimensions; d++){
				delete[] extentA[d];
				delete[] newExtentB[d];
			}
			delete[] extentA;
			delete[] newExtentB;
			extentA = newExtentA;
			NCPclusterA = NCPclusterA + incA;
			if (points[i]->inMicro) numMicroPointsA++;
			clusterA.push_back(points[i]);
		}
		else{
			for (int d=0; d<numDimensions; d++){
				delete[] extentB[d];
				delete[] newExtentA[d];
			}
			delete[] extentB;
			delete[] newExtentA;
			extentB = newExtentB;
			NCPclusterB = NCPclusterB + incB;
			if (points[i]->inMicro) numMicroPointsB++;
			clusterB.push_back(points[i]);
		}
//		cout << "done" << endl;


//		float optionA = getNCP(clusterA, points[i], theSizeWorld) - getNCP(clusterA, theSizeWorld);
//		float optionB = getNCP(clusterB, points[i], theSizeWorld) - getNCP(clusterB, theSizeWorld);
//
////		cout << "optionA=" << optionA << " optionB=" << optionB << endl;
//
//		if ( optionA < optionB ){
//			clusterA.push_back(points[i]);
//		}
//		else{
//			clusterB.push_back(points[i]);
//		}
	}
//	cout << endl;

	int leftNumPoints = clusterA.size();
	int rightNumPoints = clusterB.size();


//	if (leftNumPoints<K || rightNumPoints<K){
//		return false;
//	}

	if (leftNumPoints==0 || rightNumPoints==0){
		return false;
	}


	pointPtr* leftPoints = new pointPtr[leftNumPoints];
	pointPtr* rightPoints = new pointPtr[rightNumPoints];

	copy(clusterA.begin(), clusterA.end(), leftPoints);
	copy(clusterB.begin(), clusterB.end(), rightPoints);


	this->leafNode = false; // make this a non-leaf node!

	left = new TDNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new TDNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;

	this->left = left;
	this->right = right;

	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;

	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;


	return true;
}


float TDNode::getNCPIncrease(float** oldExtent, float** &nExtent, float oldNCP, pointPtr po, int oldNumMicroPoints,  float* sizeWorld){

//	cout << "oldExtent: " << getString(oldExtent, numDimensions) << endl;
//	cout << "adding point: " << *po << endl;
	nExtent = newExtent(oldExtent, po);
//	cout << "nExtent: " << getString(nExtent, numDimensions) << endl;

	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (nExtent[d][1] - nExtent[d][0])/sizeWorld[d];
	}

	int numMicroPoints = oldNumMicroPoints;
	if (po->inMicro) numMicroPoints++;

	float newNCP = perimeter * numMicroPoints;

	return (newNCP - oldNCP);
}


float TDNode::getNCP(vector<pointPtr> cluster, pointPtr po, float* sizeWorld){
	int numDimensions = po->numDimensions;
	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}
	vector<pointPtr>::iterator iter;

	int numMicroPoints = 0;
	for (iter=cluster.begin(); iter!=cluster.end(); iter++){
		if ((*iter)->inMicro){
			numMicroPoints++;
		}
		for (int d=0; d<numDimensions; d++){
			if (iter==cluster.begin()){
				extent[d][0] = (*iter)->p[d];
				extent[d][1] = (*iter)->p[d];
			}
			else{
				if ((*iter)->p[d]<extent[d][0]){
					extent[d][0] = (*iter)->p[d];
				}
				if ((*iter)->p[d]>extent[d][1]){
					extent[d][1] = (*iter)->p[d];
				}
			}
		}
	}

	if (po->inMicro){
		numMicroPoints++;
	}
	for (int d=0; d<numDimensions; d++){
		if (po->p[d]<extent[d][0]){
			extent[d][0] = po->p[d];
		}
		if (po->p[d]>extent[d][1]){
			extent[d][1] = po->p[d];
		}
	}

	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
	}

	return perimeter*numMicroPoints;
}

float TDNode::getNCP(vector<pointPtr> cluster, float* sizeWorld){
	int numDimensions = cluster[0]->numDimensions;
	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}
	vector<pointPtr>::iterator iter;

	int numMicroPoints = 0;
	for (iter=cluster.begin(); iter!=cluster.end(); iter++){
		if ((*iter)->inMicro){
			numMicroPoints++;
		}
		for (int d=0; d<numDimensions; d++){
			if (iter==cluster.begin()){
				extent[d][0] = (*iter)->p[d];
				extent[d][1] = (*iter)->p[d];
			}
			else{
				if ((*iter)->p[d]<extent[d][0]){
					extent[d][0] = (*iter)->p[d];
				}
				if ((*iter)->p[d]>extent[d][1]){
					extent[d][1] = (*iter)->p[d];
				}
			}
		}
	}

	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
	}

	return perimeter*numMicroPoints;
}





bool TDNode::fastSplitNode(TDNodePtr &left, TDNodePtr &right){
//	cout << "called fastSplitNode for id=" << this->id << endl;
	vector<float> temp;
	for (int i=0; i<numPoints; i++){
		temp.push_back(points[i]->p[splitDimension]);
	}

	sort(temp.begin(), temp.end()); // sort the vector!

	float median = *(temp.begin()+(temp.size()-1)/2);
	splitValue = median;

	/*******************************************/
	/* this code segment splits in median ******/
	int leftNumPoints = (temp.size()-1)/2;
	for (int i=leftNumPoints; i<numPoints; i++){
		if ( fabs(temp[i]-median)<0.001 ){
			leftNumPoints++;
		}
		else break;
	}

	int rightNumPoints = temp.size() - leftNumPoints;
//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;

	pointPtr *leftPoints, *rightPoints;

	leftPoints = new pointPtr [leftNumPoints];
	rightPoints = new pointPtr [rightNumPoints];

	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=median){ // store them in left child
//			cout << "left" << endl;
			leftPoints[leftIndex] = new point(*(points[i]));
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			rightPoints[rightIndex] = new point(*(points[i]));
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	/*******************************************/


	if (rightNumPoints==0){ // don't split
//		cerr << "**** rightNumPoints==0" << endl;
		return false; // don't split
	}

	this->leafNode = false; // make this a non-leaf node!

	left = new TDNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new TDNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;

	this->left = left;
	this->right = right;

	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;
	temp.clear();

	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;

	return true;
}



void TDNode::printNode(){
	cout << this;
}

ostream& operator<<(ostream& os, const TDNodePtr& n){
	os << "node ID=" << n->id << " leafNode=" << n->leafNode << " ";
	os << "numDimensions=" << n->numDimensions << " numPoints=" << n->numPoints << " numMicroPoints=" << n->numMicroPoints << endl;
	os << "splitDimension=" << n->splitDimension << " splitValue=" << n->splitValue << endl;
	for (int i=0; i<n->numPoints; i++){
		os << *(n->points[i]) << " ";
	}
	os << endl;
	os << "mbr:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->mbr[j][0] << " -- " << n->mbr[j][1] << " | ";
	}
	os << endl;
	os << "extent:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->extent[j][0] << " -- " << n->extent[j][1] << " | ";
	}
	os << endl;

	return os;
}


candidateSplit::candidateSplit(float splitValue, groupPtr first, groupPtr second, float* sizeWorld, int K){
	this->K = K;
	this->splitValue = splitValue;
	this->first = first;
	this->second = second;
	numDimensions = first->numDimensions;
	this->sizeWorld = sizeWorld;
}

candidateSplit::~candidateSplit(){
	delete first;
	delete second;
}

float candidateSplit::getScore(){
	int tgt1 = 1;
	while (tgt1*K<=first->points.size()) tgt1 = tgt1<<1;
	tgt1 = tgt1>>1;
//	cout << first->points.size() << " close to " << tgt1*K << endl;
	int tgt2 = 1;
	while (tgt2*K<=second->points.size()) tgt2 = tgt2<<1;
	tgt2 = tgt2>>1;
//	cout << second->points.size() << " close to " << tgt2*K << endl;

	float firstScore = first->getNormalPerimeter(sizeWorld);//* (first->points.size()%K) ;//  * (1+(first->points.size()%(K*tgt1)));
	float secondScore = second->getNormalPerimeter(sizeWorld);//* (second->points.size()%K);//  * (1+(second->points.size()%(K*tgt2)));
	return (firstScore + secondScore);//*first->points.size()*second->points.size();
//	return max(firstScore, secondScore);
}


