#include "tdtree.h"


using namespace std;

TDTree::TDTree(char *inputfile, int K){
	groupsCalculated = false;
	float *p;
	vector<float *> v;
	cout << "TDTree:: reading from " << inputfile << endl;
	fstream fin(inputfile, ios::in);

	fin >> numDimensions; // first read dimensionality
	while(!fin.eof()){
		p = new float[numDimensions];
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		v.push_back(p);
	}
	fin.close();
	cerr << "... done reading " << v.size() << " points" << endl;

	this->K = K;

//	cout << "K=" << this->K << endl;

	int numPoints = v.size();
	pointPtr *points = new pointPtr[numPoints];
	for (int i=0; i<numPoints; i++){
		points[i] = new point(numDimensions, v[i], true);
	}
	v.clear();

	root = new TDNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "TDTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;

//	cerr << "there are " << nodes.size() << " nodes" << endl;

//	cerr << "gonna do the filtering process" << endl;
//	filter(); // do the HUSTLE!
//	cerr << "...done with the filtering process" << endl;
}


TDTree::TDTree(vector<pointPtr> poiset, int K){
	groupsCalculated = false;
	numDimensions = poiset[0]->numDimensions;
	this->K = K;

	int numPoints = poiset.size();
	pointPtr *points = new pointPtr[numPoints];
	copy(poiset.begin(), poiset.end(), points);

	root = new TDNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "TDTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;

	cerr << "there are " << nodes.size() << " nodes" << endl;

//	cerr << "gonna do the filtering process" << endl;
//	filter(); // do the HUSTLE!
//	cerr << "...done with the filtering process" << endl;
}


void TDTree::recursiveSplit(TDNodePtr node){
//	cerr << "recursiveSplit @ " << node->id << endl;

	if (node->numPoints>K && node->calculateNumMicroPoints()>0){
		TDNodePtr left, right;
		bool hasSplit = node->splitNode(left, right, sizeWorld, K);

		if (hasSplit){
			nodes[left->id] = left;
			nodes[right->id] = right;

			if (left->numMicroPoints==0){
				nodes.erase(left->id);
				delete left;
				node->left = NULL;
			}
			else{
				recursiveSplit(left);
			}

			if (right->numMicroPoints==0){
				nodes.erase(right->id);
				delete right;
				node->right = NULL;
			}
			else{
				recursiveSplit(right);
			}

			if (node->left == NULL && node->right == NULL){
				node->leafNode = true;
			}

		}
	}
}


void TDTree::calculateSizeWorld(){
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}
}





void TDTree::filter(){

	list<bucket *> runts; // underfull groups
	list<bucket *> done; // ok groups

	int runtsSize = 0;
	int doneSize = 0;

	map<int, TDNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode || (iter->second->left == NULL && iter->second->right == NULL)){
			continue; // this is not a leaf node
		}

		bucket* b = new bucket();
		int numMicroPoints = 0;

		for (int i=0; i<iter->second->numPoints; i++){
			b->points.push_back(iter->second->points[i]);
			if (iter->second->points[i]->inMicro){
				numMicroPoints++;
			}
		}
		b->numMicroPoints = numMicroPoints;

		b->mbr = findExtent(b->points);

		b->node = iter->second;
		b->altered = false;

		if (b->points.size()>=K){
			done.push_back(b);
			doneSize++;
		}
		else{
			runts.push_back(b);
			runtsSize++;
		}
	}

//	cout << "runtsSize = " << runtsSize << " doneSize = " << doneSize << endl;


	while(!runts.empty()){

		bucket* runt = runts.front();
		runts.pop_front();
		runtsSize--;

//		cout << "checking new runt! " << runtsSize << " remaining" << endl;


		int thisNumPoints = runt->points.size();

		bool merge_with_done = false;

		bool gotMerge = false;
		bool gotBorrow = false;

		list <bucket *>::iterator merge_it; // to store a pointer to the bucket I merge to
        float min_merge_cost = FLT_MAX;

		list <bucket *>::iterator borrow_it; // to store a pointer to the bucket I borrow from
		float min_borrow_cost = FLT_MAX;


		// MERGE!
		list <bucket *>::iterator it;
//		cout << "gonna scan runts list" << endl;
		for (it=runts.begin(); it!=runts.end(); it++) //merge with other runt
		{
			int numPointsInUnion = (*it)->points.size() + thisNumPoints;
//			cout << "numPointsInUnion = " << numPointsInUnion << endl;

			if (numPointsInUnion < K) continue;

			float** newextent = getUnion(runt->mbr, (*it)->mbr, numDimensions);

			float NCPincrease = numPointsInUnion*getNormalizedPerimeter(newextent, sizeWorld, numDimensions) - (runt->getNCP(sizeWorld) + (*it)->getNCP(sizeWorld));
//			cout << "NCPincrease=" << NCPincrease << endl;

			for (int d=0; d<numDimensions; d++){
				delete[] newextent[d];
			}
			delete[] newextent;

			if (NCPincrease < min_merge_cost){
				min_merge_cost = NCPincrease;
				merge_it = it;
				gotMerge = true;
			}


		}
//		cout << "gonna scan done list" << endl;
		for (it=done.begin(); it!=done.end(); it++){ //merge with a done bucket
			int numPointsInUnion = (*it)->points.size() + thisNumPoints;
//			cout << "numPointsInUnion = " << numPointsInUnion << endl;

			if (numPointsInUnion < K) continue;

			float** newextent = getUnion(runt->mbr, (*it)->mbr, numDimensions);

			float NCPincrease = (runt->numMicroPoints + (*it)->numMicroPoints)*getNormalizedPerimeter(newextent, sizeWorld, numDimensions) - (runt->getNCP(sizeWorld) + (*it)->getNCP(sizeWorld));
//			cout << "NCPincrease=" << NCPincrease << endl;

			for (int d=0; d<numDimensions; d++){
				delete[] newextent[d];
			}
			delete[] newextent;

			if (NCPincrease < min_merge_cost){
				min_merge_cost = NCPincrease;
				merge_it = it;
				merge_with_done = true;
				gotMerge = true;
			}
		}


		// BORROW!
//		cout << "BORROW: gonna scan done list" << endl;
		for (it=done.begin(); it!=done.end(); it++){ // borrow from a done bucket
			int remainingNumPoints = (*it)->points.size() - (K-thisNumPoints);


			if (remainingNumPoints < 0) continue;

//			cout << "numPointsInUnion = " << numPointsInUnion << endl;

			vector<pointPtr> borrowed;
			vector<pointPtr> leftbehind;
			int leftbehindInMicro = 0;
			int borrowedInMicro = 0;


			for (uint i=0; i<(*it)->points.size(); i++){
				if (i < K-thisNumPoints){
					borrowed.push_back((*it)->points[i]);
					if ((*it)->points[i]->inMicro){
						borrowedInMicro++;
					}
				}
				else{
					leftbehind.push_back((*it)->points[i]);
					if ((*it)->points[i]->inMicro){
						leftbehindInMicro++;
					}
				}
			}


			if (leftbehindInMicro>0 && remainingNumPoints<K){ // violation of anonymity
				continue;
			}

//			cerr << "(" << borrowed.size() << "|" << leftbehind.size() << ")"; fflush(stderr);
			float** borrowedExtent = findExtent(borrowed);
			float** leftbehindExtent = findExtent(leftbehind);

			float** thisNewExtent = getUnion(borrowedExtent, runt->mbr, numDimensions);

			float newNCP = (borrowedInMicro+runt->numMicroPoints) * getNormalizedPerimeter(thisNewExtent, sizeWorld, numDimensions)
			             + leftbehindInMicro * getNormalizedPerimeter(leftbehindExtent, sizeWorld, numDimensions);

			float oldNCP = runt->getNCP(sizeWorld) + (*it)->getNCP(sizeWorld);

			if (newNCP-oldNCP < min_borrow_cost){
				min_borrow_cost = newNCP-oldNCP;
				borrow_it = it;
				gotBorrow = true;
			}

			for (int d=0; d<numDimensions; d++){
				delete[] borrowedExtent[d];
				delete[] leftbehindExtent[d];
				delete[] thisNewExtent[d];
			}
			delete[] borrowedExtent;
			delete[] leftbehindExtent;
			delete[] thisNewExtent;

		}


//		min_borrow_cost = FLT_MAX; // NEVER BORROW!


//		if (!gotMerge) cerr << "what? no merge?" << endl;
//		if (!gotBorrow) cerr << "what? no borrow" << endl;

//		cout << "minimum NCP increase is " << min_merge_cost << endl;

		if ( min_merge_cost < min_borrow_cost ){ // min_merge_cost < min_borrow_cost
//			cout << "merge option!" << endl;
			TDNodePtr other = (*merge_it)->node;
	//		cout << "ok.. "; fflush(stdout);
			// merge the runt
			for (uint i=0; i<(*merge_it)->points.size(); i++){
				runt->points.push_back( (*merge_it)->points[i] );
			}
			runt->numMicroPoints += (*merge_it)->numMicroPoints;
			for (int d=0; d<numDimensions; d++){
				delete[] runt->mbr[d];
			}
			delete[] runt->mbr;
	//		cout << "ok.. "; fflush(stdout);
	//		cout << " (" << runt->points.size() << ") "; fflush(stdout);
			runt->mbr = findExtent(runt->points);
	//		cout << "ok.. "; fflush(stdout);



			// delete the merged bucket
			if (merge_with_done){
//				delete (*merge_it);
				done.erase(merge_it);
				doneSize--;
			}
			else{
//				delete (*merge_it);
				runts.erase(merge_it);
				runtsSize--;
			}

	//		cout << "ok.. "; fflush(stdout);
			// kill the TDNode other!
			if (other->id%2==0){ // left child
				other->parent->left = NULL;
	//			if (other->parent->right == NULL){
	//				other->parent->leafNode = true;
	//			}
			}
			else{ // right child
				other->parent->right = NULL;
	//			if (other->parent->left == NULL){
	//				other->parent->leafNode = true;
	//			}
			}
	//		cout << "ok.. "; fflush(stdout);
			nodes.erase(other->id);
			delete other;
	//		cout << "ok!" << endl; fflush(stdout);

			// promote runt to done
			done.push_back(runt);
			runt->altered = true;
			doneSize++;

		}
		else { // min_merge_cost >= min_borrow_cost
//			cout << "borrow option!" << endl;
			vector<pointPtr> borrowed;
			vector<pointPtr> leftbehind;
			int leftbehindInMicro = 0;
			int borrowedInMicro = 0;
			for (int i=0; i<(*borrow_it)->points.size(); i++){
				if (i < K-thisNumPoints){
					borrowed.push_back((*borrow_it)->points[i]);
					if ((*borrow_it)->points[i]->inMicro){
						borrowedInMicro++;
					}
				}
				else{
					leftbehind.push_back((*borrow_it)->points[i]);
					if ((*borrow_it)->points[i]->inMicro){
						leftbehindInMicro++;
					}
				}
			}

			// update runt
			for (uint i=0; i<borrowed.size(); i++){
				runt->points.push_back(borrowed[i]);
			}

			runt->numMicroPoints += borrowedInMicro;
			for (int d=0; d<numDimensions; d++){
				delete[] runt->mbr[d];
			}
			delete[] runt->mbr;
			runt->mbr = findExtent(runt->points);

			// promote runt to done
			done.push_back(runt);
			runt->altered = true;
			doneSize++;



			// process what's left behind
			if (leftbehindInMicro>0){
				assert(leftbehind.size()>=K);
				(*borrow_it)->numMicroPoints = leftbehindInMicro;
				for (int d=0; d<numDimensions; d++){
					delete[] (*borrow_it)->mbr[d];
				}
				delete[] (*borrow_it)->mbr;
				(*borrow_it)->points = leftbehind;
				(*borrow_it)->mbr = findExtent((*borrow_it)->points);
				(*borrow_it)->altered = true;
			}
			else { // delete this entry and node
				TDNodePtr other = (*borrow_it)->node;
				done.erase(borrow_it);
				doneSize--;
				if (other->id%2==0){ // left child
					other->parent->left = NULL;
				}
				else{ // right child
					other->parent->right = NULL;
				}

				nodes.erase(other->id);
				delete other;
			}


		}

	}

//	cout << "there are " << doneSize << " buckets left to update" << endl;

	int affectedGroups = 0;
	list <bucket *>::iterator it;
	for (it=done.begin(); it!=done.end(); it++){
		if ( (*it)->altered ) {
			affectedGroups++;
			TDNodePtr thenode = (*it)->node;
			delete[] thenode->points;
			thenode->numPoints = (*it)->points.size();
			thenode->points = new pointPtr[thenode->numPoints];
			copy( (*it)->points.begin(), (*it)->points.end(), thenode->points);
			for (int d=0; d<numDimensions; d++){
				delete[] thenode->mbr[d];
				delete[] thenode->extent[d];
			}
			delete[] thenode->mbr;
			delete[] thenode->extent;
			thenode->calculateExtentMBR();
			thenode->calculateNumMicroPoints();
		}
	}

//	cout << "there were " << affectedGroups << " affected buckets" << endl;

//	cout << "exiting filter" << endl;

}


vector<group> TDTree::getGroups(){
	if (!groupsCalculated){
		groups.clear();
		groupsCalculated = true;

		map<int, TDNodePtr>::iterator iter;
		for (iter=nodes.begin(); iter!=nodes.end(); iter++){
			if (!iter->second->leafNode){
				continue; // this is not a leaf node
			}
			group* gp = new group(numDimensions);

			for (int i=0; i<iter->second->numPoints; i++){
				gp->insertPoint(*(iter->second->points[i]));
			}
			groups.push_back(*gp);
		}
	}
	return groups;
}

void TDTree::writeGroups(char *inputfile){
	ostringstream kdfied;
	kdfied << "K" << K << "_" << inputfile;
	string writefile = kdfied.str();
	ofstream fout(writefile.c_str(), ios::out);

	map<int, TDNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		fout << iter->second << endl;
	}

	fout.close();
}


void TDTree::printMetrics(){
	map<int, TDNodePtr>::iterator iter;

	float normalPerimeter = 0;
	float sizeTimesPerimeter = 0;
	float discernability = 0;
	int numGroups = 0;

	float sizeWorld[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}

	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		numGroups++;
		discernability += iter->second->numMicroPoints*iter->second->numMicroPoints;

		float thisPerimeter = 0;
		for (int i=0; i<numDimensions; i++){
//			cout << "... " << (iter->second->extent[i][1] - iter->second->extent[i][0])/sizeWorld[i] << endl;
			float diff = iter->second->extent[i][1] - iter->second->extent[i][0];
			thisPerimeter += diff/sizeWorld[i];
		}
		normalPerimeter += thisPerimeter;
		sizeTimesPerimeter += thisPerimeter*iter->second->numMicroPoints;
	}

	cout << "there are " << numGroups << " groups" << endl;
	cout << "discernability = " << discernability << endl;
//	cout << "normalized average equivalence class size = " << ((float) K*nodes[1]->numMicroPoints)/numGroups << endl;
	cout << "total perimeter = " << normalPerimeter << endl;
	cout << "total (perimeter)*(numMicroPoint) = " << sizeTimesPerimeter << endl;

}


void TDTree::boostWithPublicData(TDNodePtr node, KDTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (!node->leafNode){
		if (node->left!=NULL){
			boostWithPublicData(node->left, pub);
		}
		if (node->right!=NULL){
			boostWithPublicData(node->right, pub);
		}
		if (node->left == NULL && node->right == NULL){
			cerr << "both children NULL but not marked as leaf" << endl;
		}
	}
	else{ // do the range query
		node->boostedLeafNode = true;
//		cout << "do the range query" << endl;
		node->calculateExtentMBR();
//		cout << "range: " << getString(node->mbr, numDimensions) << endl;
		vector<pointPtr> newPoints = pub->rangeQuery(node->mbr);
//		cout << "query ok" << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
			(*resIter)->inMicro = false;
		}

		vector<pointPtr> thesePoints(node->numPoints);
		copy(node->points, node->points+node->numPoints, thesePoints.begin());

//		cout << "newPoints.size()=" << newPoints.size() << endl;
//		for (uint i=0; i<newPoints.size(); i++){
//			cout << *(newPoints[i]) << " ";
//		}
//		cout << endl;

//		cout << "thesePoints.size()=" << thesePoints.size() << endl;
//		for (uint i=0; i<thesePoints.size(); i++){
//			cout << *(thesePoints[i]) << " ";
//		}
//		cout << endl;

		vector<pointPtr> joinedPoints = SortMerge(newPoints, thesePoints);

//		cout << "joinedPoints.size()=" << joinedPoints.size() << endl;

		delete[] node->points;

		node->numPoints = joinedPoints.size();
		node->points = new pointPtr[node->numPoints];
		copy(joinedPoints.begin(), joinedPoints.end(), node->points);
		node->calculateNumMicroPoints();
//		cout << "numMicroPoints=" << node->numMicroPoints << endl;
		recursiveSplit(node);

	}
}




void TDTree::boostWithPublicData(TDTreePtr pub){
	groupsCalculated = false;
	map<int, TDNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (iter->second->leafNode){
			iter->second->boostedLeafNode = true;
		}
	}

//	int lStart = -1;
//	int lEnd = -1;
////	map<int, TDNodePtr>::iterator iter;
//	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
//		if (!iter->second->leafNode){
//			continue; // this is not a leaf node
//		}
//		if (lStart==-1) lStart = iter->first;
//		lEnd = iter->first;
//	}


	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->boostedLeafNode){
			continue; // this is not a leaf node that is going to be boosted
		}

		vector<pointPtr> result = pub->rangeQuery(iter->second->mbr);
		vector<pointPtr>::iterator resIter;
		for (resIter=result.begin(); resIter!=result.end(); resIter++){
			(*resIter)->inMicro = false;
		}
//		cout << "range:: | ";
//		for (int j=0; j<numDimensions; j++){
//			cout << iter->second->extent[j][0] << " -- " << iter->second->extent[j][1] << " | ";
//		}
//		cout << endl;
//		cout << "for node " << iter->second->id << " rangeQueryStr returned " << result.size() << " results" << endl;
//		for (uint i=0; i<result.size(); i++){
//			cout << *(result[i]) << " ";
//		}
//		cout << endl;

		for (int i=0; i<iter->second->numPoints; i++){
			pointPtr mtp = iter->second->points[i];
			for (resIter=result.begin(); resIter!=result.end(); resIter++){
				if ( *mtp == *(*resIter) ){
					(*resIter)->inMicro = true;
					break;
				}
			}
		}

		pointPtr *newpoints = new pointPtr[result.size()];
		for (uint i=0; i<result.size(); i++){
			newpoints[i] = new point(*(result[i]));
		}

		for (int i=0; i<iter->second->numPoints; i++){
			delete iter->second->points[i];
		}
		delete[] iter->second->points;

		iter->second->points = newpoints;
		iter->second->numMicroPoints = iter->second->numPoints;
		iter->second->numPoints = result.size();

		recursiveSplit(iter->second);

	}

}


vector<pointPtr> TDTree::rangeQuery(float **query){
	vector<pointPtr> result;

	recRangeQuery(root, query, result);

	return result;
}

void TDTree::recRangeQuery(TDNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << endl;
	if ( node->left==NULL && node->right==NULL ){ // leaf node
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
			if ( inside(node->points[i]->p, query, numDimensions) ){ // point inside
				res.push_back(node->points[i]);
//				cout << "in " << getString(node->points[i], numDimensions) << endl;
			}
			else{
//				cout << "out " << getString(node->points[i], numDimensions) << endl;
			}
		}
	}
	if ( node->left!=NULL && overlaps(query, node->left->extent, numDimensions) ){
		recRangeQuery(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlaps(query, node->right->extent, numDimensions) ){
		recRangeQuery(node->right, query, res); // descend
	}

}

vector<pointPtr> TDTree::rangeQueryStr(float **query){
	vector<pointPtr> result;

	recRangeQueryStr(root, query, result);

	return result;
}

void TDTree::recRangeQueryStr(TDNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << endl;
	if ( node->left==NULL && node->right==NULL ){ // leaf node
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
			if ( insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				res.push_back(node->points[i]);
//				cout << "in " << getString(node->points[i], numDimensions) << endl;
			}
			else{
//				cout << "out " << getString(node->points[i], numDimensions) << endl;
			}
		}
	}
	if ( node->left!=NULL && overlapsStr(query, node->left->extent, numDimensions) ){
		recRangeQueryStr(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlapsStr(query, node->right->extent, numDimensions) ){
		recRangeQueryStr(node->right, query, res); // descend
	}

}



void TDTree::markLeaves(){
	recMarkLeaves(root);
}

void TDTree::recMarkLeaves(TDNodePtr node){
	if (node->left == NULL && node->right == NULL){
		node->leafNode = true;
	}
	else {
		node->leafNode = false;
		if (node->left != NULL){
			recMarkLeaves(node->left);
		}
		if (node->right != NULL){
			recMarkLeaves(node->right);
		}
	}
}



void TDTree::countQuery(float **query, double& estimate, double& actual){
	recCountQuery(root, query, estimate, actual);
}

void TDTree::recCountQuery(TDNodePtr node, float **query, double& estimate, double& actual){
//	cout << "recCountQuery @ " << node->id << endl;

	if (node->leafNode){
		float **inter;
		inter = intersect(query, node->mbr, numDimensions);
		double interVol = 1, nodeVol = 1;
		for (int i=0; i<numDimensions; i++){
			interVol *= inter[i][1] - inter[i][0];
			nodeVol *= node->mbr[i][1] - node->mbr[i][0];
			delete inter[i];
		}
		delete inter;
		if (interVol==0) interVol = 1;
		if (nodeVol==0) nodeVol = 1;
//		cout << "interVol=" << interVol << " nodeVol=" << nodeVol << endl;
//		cout << "estimated MT tuples " << (interVol/nodeVol) * node->numMicroPoints << endl;
		estimate += (interVol/nodeVol) * node->numMicroPoints;


		double numMicro = 0;
		for (int i=0; i<node->numPoints; i++){
			if (node->points[i]->inMicro &&  insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				numMicro++;
			}
		}
//		cout << "actual MT tuples " << numMicro << endl;


		actual += numMicro;

	}
	else {
		if (node->left != NULL  &&  overlaps(query, node->left->mbr, numDimensions)){
			recCountQuery(node->left, query, estimate, actual);
		}
		if (node->right != NULL  &&  overlaps(query, node->right->mbr, numDimensions)){
			recCountQuery(node->right, query, estimate, actual);
		}
	}

}




void TDTree::countQueryREF(float **query, double& estimate, double& actual){
	recCountQueryREF(root, query, estimate, actual);
}


void TDTree::recCountQueryREF(TDNodePtr node, float **query, double& estimate, double& actual){
//	cout << "recCountQueryREF @ " << node->id << endl;

	if (node->leafNode){
		float **inter;
		inter = intersect(query, node->mbr, numDimensions);
		double interVol = 1, nodeVol = 1;
		for (int i=0; i<numDimensions; i++){
			interVol *= inter[i][1] - inter[i][0];
			nodeVol *= node->mbr[i][1] - node->mbr[i][0];
			delete inter[i];
		}
		delete inter;
		if (interVol==0) interVol = 1;
		if (nodeVol==0) nodeVol = 1;
//		cout << "interVol=" << interVol << " nodeVol=" << nodeVol << endl;
//		cout << "estimated MT tuples " << (interVol/nodeVol) * node->perceivedMicroTuples << endl;
		estimate += (interVol/nodeVol) * node->perceivedMicroTuples;


		double numMicro = 0;
		for (int i=0; i<node->numPoints; i++){
			if (node->points[i]->inMicro &&  insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				numMicro++;
			}
		}
//		cout << "actual MT tuples " << numMicro << endl;


		actual += numMicro;

	}
	else {
		if (node->left != NULL  &&  overlaps(query, node->left->mbr, numDimensions)){
			recCountQueryREF(node->left, query, estimate, actual);
		}
		if (node->right != NULL  &&  overlaps(query, node->right->mbr, numDimensions)){
			recCountQueryREF(node->right, query, estimate, actual);
		}
	}

}



void TDTree::calculatePerceivedMicroTuples(){
	recCalculatePerceivedMicroTuples(root, 1);
}

void TDTree::recCalculatePerceivedMicroTuples(TDNodePtr node, double ratio){
//	cout << "recCalculatePerceivedMicroTuples @ " << node->id << endl;

//	if (node->parent != NULL && node->numMicroPoints < K && node->parent->numMicroPoints >= K){
//		ratio = ((double) node->parent->numMicroPoints) / node->parent->leafTuples;
//	}


	if (node->boostedLeafNode){
		ratio = ((double) node->numMicroPoints) / node->leafTuples;
	}


//	cout << "ratio is " << ratio << endl;

	if (node->leafNode){
		node->perceivedMicroTuples = ratio * node->leafTuples;
//		cout << "leaf node " << node->id << " has perceivedMicroTuples=" << node->perceivedMicroTuples << " leafTuples=" << node->leafTuples << " and numMicroPoints=" << node->numMicroPoints<< endl;
	}
	else {
		if (node->left != NULL){
			recCalculatePerceivedMicroTuples(node->left, ratio);
		}
		if (node->right != NULL){
			recCalculatePerceivedMicroTuples(node->right, ratio);
		}
	}
}


void TDTree::countPerceivedMicroTuples(){
	double count = 0, actual = 0;
	recCountPerceivedMicroTuples(root, count, actual);
	cerr << "countPerceivedMicroTuples is " << count << endl;
	cerr << "actual MicroTuples is " << actual << endl;
}

void TDTree::recCountPerceivedMicroTuples(TDNodePtr node, double& count, double& actual){
//	cout << "recCountPerceivedMicroTuples @ " << node->id << endl;

	if (node->leafNode){
		count += node->perceivedMicroTuples;
		actual += node->numMicroPoints;
//		cout << " + " << node->perceivedMicroTuples << " tuples" << endl;
	}
	else {
		if (node->left != NULL){
			recCountPerceivedMicroTuples(node->left, count, actual);
		}
		if (node->right != NULL){
			recCountPerceivedMicroTuples(node->right, count, actual);
		}
	}
}



void TDTree::calculateLeafTuples(){
	recCalculateLeafTuples(root);
	cout << "root has " << root->leafTuples << " tuples" << endl;
}


void TDTree::recCalculateLeafTuples(TDNodePtr node){
//	cout << "recCalculateLeafTuples @ " << node->id << endl;
	if (node->leafNode){
//		cout << "node->numMicroPoints = " << node->numMicroPoints << endl;

		if (node->numMicroPoints >= K){
			node->leafTuples = node->numMicroPoints;
		}
		else {
			node->leafTuples = K;
		}
//		cout << "leaf node " << node->id << " has tuples " << node->leafTuples << endl;
	}
	else {
		if (node->left == NULL && node->right == NULL){
			cerr << "both children NULL" << endl;
		}
		node->leafTuples = 0;
		if (node->left != NULL){
			recCalculateLeafTuples(node->left);
			node->leafTuples += node->left->leafTuples;
		}
		if (node->right != NULL){
			recCalculateLeafTuples(node->right);
			node->leafTuples += node->right->leafTuples;
		}
//		cout << "--- node " << node->id << " has tuples " << node->leafTuples << endl;
	}
}


void TDTree::calculateMicroTuples(){
	double numMicroTuples = 0;
	recCalculateMicroTuples(root, numMicroTuples);
	cout << "root has " << numMicroTuples << " micro tuples" << endl;
}


void TDTree::recCalculateMicroTuples(TDNodePtr node, double& numMicroTuples){
	if (node->leafNode){
		numMicroTuples += node->numMicroPoints;
	}
	else {
		node->leafTuples = 0;
		if (node->left != NULL){
			recCalculateMicroTuples(node->left, numMicroTuples);
		}
		if (node->right != NULL){
			recCalculateMicroTuples(node->right, numMicroTuples);
		}
	}
}



ostream& operator<<(ostream& os, const TDTreePtr& t){
	map<int, TDNodePtr>::iterator iter;

	cout << endl;
	for(iter = t->nodes.begin(); iter!=t->nodes.end(); iter++){
        cout << "map id=" << iter->first << " " << iter->second << endl;
    }

	return os;
}

